"""
Repositories which give an overiew of statuses etc
Created by adam on 3/14/20
"""
__author__ = 'adam'

from CanvasHacks import environment
from CanvasHacks.DAOs.mixins import DaoMixin
from CanvasHacks.Definitions.skaa import InitialWork,\
    MetaReview, Review
from CanvasHacks.Definitions.discussion import DiscussionForum, DiscussionReview
from CanvasHacks.Definitions.groups import SkaaReviewGroup, DiscussionGroup
from CanvasHacks.Repositories.reviewer_associations import AssociationRepository
from CanvasHacks.Repositories.status import FeedbackStatusRepository, InvitationStatusRepository
from CanvasHacks.Repositories.students import StudentRepository

if __name__ == '__main__':
    pass
import pandas as pd


class SkaaOverviewRepository( DaoMixin ):
    """
    Holds consolidated information about statuses etc for
    Skaa assignments

    Other classes like dashboard.skaaDashboard are in
    charge of displaying the information
    """

    def __init__( self, unit=None ):
        self.unit = unit
        self.studentRepo = StudentRepository( environment.CONFIG.course )
        self.studentRepo.download()

        if unit is not None:
            # Load and display counts
            self.load( unit )

    def _initialize( self, unit ):
        self.data = [ ]

        self.unit = unit
        # The activity whose id is used to store review pairings for the whole SKAA
        self.activity_for_review_pairings = self.unit.initial_work

        self.components = [ c for c in self.unit.components if isinstance( c, SkaaReviewGroup ) ]


        self._initialize_db()

        self.assignRepo = AssociationRepository( self.dao, self.activity_for_review_pairings )

    def load( self, unit ):
        """
        Main called method which initializes and loads data
        Often will have to be called later than initialization
        so that we can have the object existing and then specify
        which unit to load.

        :param unit:
        :return:
        """
        self._initialize( unit )

        for sid, obj in self.studentRepo.data.items():
            d = {
                'student': obj.name,
                'canvas_id': sid,
                'csun_id': obj.sis_user_id
            }
            for c in self.components:
                if len( self.assignRepo.get_associations() ) > 0:
                    try:
                        # get the record where the student is the reviwer
                        a = self.assignRepo.get_by_reviewer( sid )
                        # get the name of the student being assessed
                        d[ 'reviewing' ] = self.studentRepo.get_student_name( a.assessee_id )
                        d[ 'reviewing_id' ] = a.assessee_id
                        # get the record where the student is the author
                        b = self.assignRepo.get_by_author( sid )
                        # get the name
                        d[ 'reviewed_by' ] = self.studentRepo.get_student_name( b.assessor_id )
                        d[ 'reviewed_by_id' ] = b.assessor_id
                    except AttributeError:
                        pass

                self.add_invites( d, c, sid )

                self.add_reviews( d, c, sid )

            self.data.append( d )

        self.data = pd.DataFrame( self.data )

    def add_invites( self, data_dict, component, student_id ):
        invite_fields = { Review: 'invited_to_review', MetaReview: 'invited_to_metareview',
                          DiscussionReview: 'invited_to_discussion_review' }

        invite_fieldname = invite_fields.get( type( component ) )

        if invite_fieldname is not None:
            inv = InvitationStatusRepository( self.dao, component )
            data_dict[ invite_fieldname ] = pd.to_datetime( inv.received_at( student_id ) )

    def add_reviews( self, data_dict, component, student_id ):
        # Note: can't do in similar way to invitations since invited to metareview and received ca feedback
        # use different activities. The invitation is for the upcoming one which provides feedback

        # on the previous one
        # set to none so won't overwrite on next time through
        fb_fieldname = None

        if isinstance( component, InitialWork ):
            # we can't use the review object because feedback on the review
            # comes from the metareview
            fb_fieldname = 'received_feedback_on_essay'

        if isinstance( component, Review ):
            fb_fieldname = 'received_feedback_on_review'

        if isinstance( component, DiscussionForum ):
            fb_fieldname = 'received_discussion_feedback'

        if fb_fieldname is not None:
            fr = FeedbackStatusRepository( self.dao, component )
            data_dict[ fb_fieldname ] = pd.to_datetime( fr.received_at( student_id ) )

    @property
    def essay( self ):
        """
        Return students who have done initial work and been assigned a reviewer

        :return: DataFrame
        """
        return self.data[ ~self.data.reviewing.isnull() ]

    @property
    def no_essay( self ):
        """
        Students who have not submitted the initial work
        :return: DataFrame
        """
        return self.data[ self.data.reviewing.isnull() ]

    @property
    def reviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose reviewer has turned in the review

        :return: DataFrame
        """
        # Students whose reviewer has and has not turned in review
        return self.essay[ ~self.essay.received_feedback_on_essay.isnull() ]

    @property
    def non_reviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose reviewer has NOT turned in the review

        :return: DataFrame
        """
        return self.essay[ self.essay.received_feedback_on_essay.isnull() ]\
            # .drop( [ 'reviewing' ], axis=1 )

    @property
    def metareviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose author has turned in the metareview

        :return: DataFrame
        """
        return self.essay[ ~self.essay.received_feedback_on_review.isnull() ]

    @property
    def non_metareviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose author has turned in the metareview

        :return: DataFrame
        """
        return self.essay[ self.essay.received_feedback_on_review.isnull() ]


class DiscussionOverviewRepository( DaoMixin ):
    """
    Holds information about the discussion and discussion
    review for a unit

    Other classes like dashboard.discussionDashboard are in
    charge of displaying the information
    """

    def __init__( self, unit=None ):
        """
        Initializes and loads all data or
        waits to have load called

        :return:
        """
        self.unit = unit
        self.studentRepo = StudentRepository( environment.CONFIG.course )
        self.studentRepo.download()

        if unit is not None:
            # Load and display counts
            self.load( unit )

    def _initialize( self, unit ):
        self.data = [ ]
        self.unit = unit

        # The activity whose id is used to store review pairings for the whole SKAA
        self.activity_for_review_pairings = self.unit.discussion_review
        self.components = [ c for c in self.unit.components if isinstance( c, DiscussionGroup ) ]
        self.studentRepo = StudentRepository( environment.CONFIG.course )
        self.studentRepo.download()
        self._initialize_db()
        self.assignRepo = AssociationRepository( self.dao, self.activity_for_review_pairings )

    def load( self, unit ):

        self._initialize( unit )

        for sid, obj in self.studentRepo.data.items():
            d = {
                'student': obj.name,
                'canvas_id': sid,
                'csun_id': obj.sis_user_id
            }
            for c in self.components:
                if len( self.assignRepo.get_associations() ) > 0:
                    try:
                        # get the record where the student is the reviwer
                        a = self.assignRepo.get_by_reviewer( sid )
                        # get the name of the student being assessed
                        d[ 'reviewing' ] = self.studentRepo.get_student_name( a.assessee_id )
                        d[ 'reviewing_id' ] = a.assessee_id
                        # get the record where the student is the author
                        b = self.assignRepo.get_by_author( sid )
                        # get the name
                        d[ 'reviewed_by' ] = self.studentRepo.get_student_name( b.assessor_id )
                        d[ 'reviewed_by_id' ] = b.assessor_id
                    except AttributeError:
                        pass

                self.add_invites( d, c, sid )

                self.add_reviews( d, c, sid )

            self.data.append( d )

        self.data = pd.DataFrame( self.data )

    def add_invites( self, data_dict, component, student_id ):
        invite_fields = { DiscussionReview: 'invited_to_discussion_review' }

        invite_fieldname = invite_fields.get( type( component ) )

        if invite_fieldname is not None:
            inv = InvitationStatusRepository( self.dao, component )
            data_dict[ invite_fieldname ] = pd.to_datetime( inv.received_at( student_id ) )

    def add_reviews( self, data_dict, component, student_id ):
        # Note: can't do in similar way to invitations since invited to metareview and received ca feedback
        # use different activities. The invitation is for the upcoming one which provides feedback
        # on the previous one

        # set to none so won't overwrite on next time through
        fb_fieldname = None

        if isinstance( component, DiscussionForum ):
            fb_fieldname = 'received_discussion_feedback'

        if fb_fieldname is not None:
            fr = FeedbackStatusRepository( self.dao, component )
            data_dict[ fb_fieldname ] = pd.to_datetime( fr.received_at( student_id ) )

    @property
    def posters( self ):
        """
        Students who have posted and been assigned reviewers
        todo Consider whether should be using the assignment of reviewers or a status object
        :return: DataFrame
        """
        return self.data[ ~self.data.reviewing.isnull() ]

    @property
    def non_posters( self ):
        """
        Students who have not posted and thus not been assigned a reviewer
        :return: DataFrame
        """
        return self.data[ self.data.reviewing.isnull() ]

    @property
    def reviewed( self ):
        """
        Students whose reviewer has turned in the review
        :return: DataFrame
        """
        return self.posters[ ~self.posters.received_discussion_feedback.isnull() ]

    @property
    def non_reviewed( self ):
        """
        Students whose reviewer has NOT turned in the review
        :return: DataFrame
        """
        return self.posters[ self.posters.received_discussion_feedback.isnull() ]

    def add_invites( self, data_dict, component, student_id ):
        invite_fields = { DiscussionReview: 'invited_to_discussion_review' }

        invite_fieldname = invite_fields.get( type( component ) )

        if invite_fieldname is not None:
            inv = InvitationStatusRepository( self.dao, component )
            data_dict[ invite_fieldname ] = pd.to_datetime( inv.received_at( student_id ) )

    def add_reviews( self, data_dict, component, student_id ):
        # Note: can't do in similar way to invitations since invited to metareview and received ca feedback
        # use different activities. The invitation is for the upcoming one which provides feedback
        # on the previous one

        # set to none so won't overwrite on next time through
        fb_fieldname = None

        # todo this probably should be changed to discussion forum everywhere
        if isinstance( component, DiscussionReview ):
            fb_fieldname = 'received_discussion_feedback'

        if fb_fieldname is not None:
            fr = FeedbackStatusRepository( self.dao, component )
            data_dict[ fb_fieldname ] = pd.to_datetime( fr.received_at( student_id ) )


class NonSkaaAssignmentOverviewRepository:

    def __init__( self ):
        pass



import unittest
from unittest.mock import create_autospec, MagicMock

from CanvasHacks.Processors.filters import SubmissionReadinessFilter, FrameRowReadinessFilter, ReadinessFilterFactory,\
    QuizSubmissionReadinessFilter, UncheckableDataProvidedError
from canvasapi.quiz import QuizSubmission
from canvasapi.submission import Submission
import pandas as pd

from unittest import TestCase

from TestingBase import TestingBase
if __name__ == '__main__':
    pass


class TestSubmissionReadinessFilter(TestingBase):

    def setUp(self):
        self.config_for_test()
        self.obj = SubmissionReadinessFilter()

    def test_is_ready( self ):
        t = MagicMock(grade='complete')
        self.assertFalse(self.obj.is_ready(t), "returns false when grade is complete")

        t = MagicMock(workflow_state='complete')
        self.assertFalse(self.obj.is_ready(t), "returns false when workflow state is unsubmitted")

        t = MagicMock(workflow_state='pending_review')
        self.assertTrue(self.obj.is_ready(t), "returns true when workflow state is pending_review")

    def test_make_raises_on_unknown( self ):
        with self.assertRaises(UncheckableDataProvidedError):
            t = create_autospec( Submission )
            self.obj.is_ready(t)



class TestFrameRowReadinessFilter( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.obj = FrameRowReadinessFilter()

    def test_is_ready( self ):
        self.skipTest( 'todo' )



class TestQuizSubmissionReadinessFilter( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.obj = QuizSubmissionReadinessFilter()

    def test_is_ready( self ):
        t = create_autospec(QuizSubmission)
        self.skipTest( 'todo' )


class TestReadinessFilterFactory( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        ReadinessFilterFactory()

    def test_make( self ):
        i = pd.Series()
        f = ReadinessFilterFactory.make(i)
        self.assertIsInstance(f, FrameRowReadinessFilter, "works for FrameRowReadinessFilter")

        i = create_autospec(Submission)
        f = ReadinessFilterFactory.make(i)
        self.assertIsInstance(f, SubmissionReadinessFilter, "works for SubmissionReadinessFilter")

        i = create_autospec(QuizSubmission)
        f = ReadinessFilterFactory.make( i )
        self.assertIsInstance( f, QuizSubmissionReadinessFilter, "works for QuizSubmissionReadinessFilter" )

    def test_make_raises_on_unknown( self ):
        with self.assertRaises(UncheckableDataProvidedError): #,  "raises when given string" ):
            f = ReadinessFilterFactory.make( "taco" )


if __name__ == '__main__':
    unittest.main()



"""
Created by adam on 3/2/20
"""
__author__ = 'adam'

from unittest.mock import MagicMock, create_autospec, patch

from faker import Faker

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Messaging.discussions import FeedbackFromDiscussionReviewMessenger
from CanvasHacks.Definitions.discussion import DiscussionReview
from CanvasHacks.Repositories.status import FeedbackStatusRepository
from CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster import SendDiscussionReviewToPoster
from tests.TestingBase import TestingBase
from tests.factories.ModelFactories import student_factory
from tests.factories.PeerReviewedFactories import unit_factory
from tests.factories.RepositoryMocks import ContentRepositoryMock

fake = Faker()

if __name__ == '__main__':
    pass


class TestUnitTests(TestingBase):
    """Unit tests with all dependencies mocked"""
    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.activity = self.unit.discussion_review
        self.course = MagicMock()

        # student recieiving the message
        self.author = student_factory()
        self.reviewer = student_factory()

        # This would be the content unit
        self.work = fake.text()
        # Set to none so that loader thinks not a quiz
        self.activity_id = self.unit.discussion_review.id
        self.dao = SqliteDAO()
        self.session = self.dao.session
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects

        # self.studentRepo = StudentRepository()
        # self.studentRepo.get_student = MagicMock( return_value=self.reviewer )
        self.contentRepo = ContentRepositoryMock()
        self.contentRepo.get_formatted_work_by = MagicMock( return_value=self.work )
        self.review_assign = MagicMock( assessor_id=self.reviewer.id, assessee_id=self.author.id )
        self.statusRepo = MagicMock()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data( self.student_ids, make_dataframe=True )

    def test_instantiates_correct_status_repos( self ):
        """The sending of metareview results requires a
        special status repository"""
        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )

        self.assertIsInstance(obj.statusRepos, list, "status repos is a list")
        self.assertTrue(len(obj.statusRepos) == 1)
        self.assertIsInstance(obj.feedback_status_repo, FeedbackStatusRepository, "Invite repo instantiated")

    @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.FeedbackStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.WorkRepositoryLoaderFactory' )
    def test__message_step( self, workLoaderMock, studentRepoMock, messengerMock, statusRepoMock ):
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )

        # setup work repository (use all students since no filtering occurs here)
        self.workRepo.submitter_ids = self.student_ids
        obj.work_repo = self.workRepo
        # setup students
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )
        # setup review pairings
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.students,
                                                                        obj.dao.session )
        obj.associations = preexisting_pairings

        # call
        obj._message_step()

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.student_ids ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]
        # print(args)

        # Check that all messages have the correct subject
        for sid, subj, body in messenger_args:
            expect_subj = FeedbackFromDiscussionReviewMessenger.email_subject_templ.format(self.unit.unit_number)
            self.assertEqual( expect_subj, subj, "Correct subject line" )

        # Status repo calls on messenger
        obj.messenger.status_repositories[0].record.assert_called()
        call_list = obj.messenger.status_repositories[0].record.call_args_list
        status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        self.assertEqual( len( self.students ), len( call_list ),
                          "Status repo record called expected number of times" )
        for sid in self.student_ids:
            self.assertIn( sid, status_args, "StatusRepo.record called on all students" )

        # student repo calls on messenger
        for sid in self.student_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # This is the review which the assessor has submitted
            author_text = self.workRepo.get_formatted_work_by( record.assessor_id )
            # which we are sending to the person who created the posts
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( author_text )
            self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.DisplayManager' )
    def test__assign_step( self, displayManagerMock ):
        # self.workRepo.remove_student_records = MagicMock()
        # workLoaderMock.make = MagicMock( return_value=self.workRepo )

        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.work_repo = MagicMock( submitter_ids=self.preexisting_student_ids )
        obj._filter_notified = MagicMock()
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.students, obj.dao.session )

        # call
        obj._assign_step()

        # check
        obj._filter_notified.assert_called()
        self.assertEqual( len( self.preexisting_student_ids ), len( obj.associations ) )

    def test__filter_notified( self ):
        num_previous = 2
        # Instantiate so can use its dao for setup
        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )
        self.create_preexisting_review_pairings( self.activity_id, self.students, obj.dao.session )

        self.make_feedback_received_records( num_previous, session=obj.dao.session )

        self.workRepo = MagicMock()
        obj.work_repo = self.workRepo

        reviewers_with_notified_authors = [ r.assessor_id for r in self.pairings if r.assessee_id in self.previously_sent ]

        # print(reviewers_with_authors_sent_feedback)
        # ra = obj.dao.session.query(ReviewAssociation).all()
        # fb = obj.dao.session.query(FeedbackReceivedRecord).all()

        # call
        obj._filter_notified()

        # check
        self.assertIsInstance( obj.statusRepos[0], FeedbackStatusRepository, "Correct status repo used" )

        self.workRepo.remove_student_records.assert_called()

        call_args = [ c[ 0 ][ 0 ] for c in self.workRepo.remove_student_records.call_args_list ]

        print(call_args)

        expected = [r for r in self.student_ids if r in reviewers_with_notified_authors]
        self.assertEqual( expected, call_args[ 0 ],
                          "Submitters whose authors have already been sent feedback were removed" )

        #
        #
        # num_previously_sent_results = 3
        # previously_notified_authors = self.student_ids[ :num_previously_sent_results ]
        # non_notified_authors = self.student_ids[ num_previously_sent_results: ]
        # self.assertEqual( len( self.student_ids ),
        #                   len( non_notified_authors ) + len( previously_notified_authors ),
        #                   "dummy check of setup" )
        #
        # # Instantiate so can use its dao for setup
        # obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )
        # self.create_preexisting_review_pairings( self.activity_id, self.students, obj.dao.session )
        #
        # for sid in non_notified_authors:
        #     obj.statusRepos.record_sent_feedback(sid)
        #
        # assocs = obj.dao.session.query( ReviewAssociation )\
        #     .filter( ReviewAssociation.activity_id == self.activity_id) \
        #     .all()
        #
        # submitters_whose_authors_have_been_notified = [ s for s in assocs if s.assessee_id in previously_notified_authors ]
        #
        # # submitters_whose_authors_have_been_notified = [ obj.associationRepo.get_by_assessee( self.activity, sid ) for sid in previously_notified_authors ]
        #
        # submitters_whose_authors_have_been_notified = [ s.assessor_id for s in
        #                                                 submitters_whose_authors_have_been_notified ]
        #
        # submitters_whose_authors_not_notified = [ s for s in self.student_ids if
        #                                           s not in submitters_whose_authors_have_been_notified ]
        #
        # # Suppose that everyone has submitted their review before this run
        # self.workRepo.submitter_ids = self.student_ids
        #
        # # obj.statusRepos = create_autospec( FeedbackStatusRepository,
        # #                                               previously_received_feedback=non_notified_authors )
        #
        # self.workRepo.remove_student_records = MagicMock()
        # obj.work_repo = self.workRepo
        #
        # # call
        # obj._filter_notified()


    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.DisplayManager' )
    @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.WorkRepositoryLoaderFactory' )
    def test__load_step( self, workLoaderMock, displayManagerMock ):
        self.workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )
        # obj.display_manager = create_autospec(DisplayManager)

        # obj.statusRepos = create_autospec(FeedbackStatusRepository,  previously_sent_students=self.preexisting_student_ids)
        # call
        obj._load_step()

        # check
        workLoaderMock.make.assert_called()
        displayManagerMock.assert_called()

        call_args = [ c[ 0 ][ 0 ] for c in workLoaderMock.call_args_list ]
        print( call_args )
        # self.assertEqual()

        # obj.display_manager.assert_called()

        # self.workRepo.remove_student_records.assert_called()
        # call_args = [  c[ 0 ][ 0 ] for c in self.workRepo.remove_student_records.call_args_list]
        # self.assertEqual(self.preexisting_student_ids, call_args[0] )

    def test_audit_frame( self ):
        self.skipTest( "todo" )


class TestFunctionalTests( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.activity = self.unit.discussion_review
        self.course = MagicMock()

        # student recieiving the message
        self.author = student_factory()
        self.reviewer = student_factory()

        # This would be the content unit
        self.work = fake.text()
        # Set to none so that loader thinks not a quiz
        self.activity_id = self.unit.discussion_review.id
        self.dao = SqliteDAO()
        self.session = self.dao.session
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects

        # self.studentRepo = StudentRepository()
        # self.studentRepo.get_student = MagicMock( return_value=self.reviewer )
        self.contentRepo = ContentRepositoryMock()
        self.contentRepo.get_formatted_work_by = MagicMock( return_value=self.work )
        self.review_assign = MagicMock( assessor_id=self.reviewer.id, assessee_id=self.author.id )
        self.statusRepo = MagicMock()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data( self.student_ids, make_dataframe=True )

    def test_instantiates_w_correct_values( self ):
        """Dummy checking in case some of the main variables used change"""
        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )

        self.assertIsInstance( obj.statusRepos[0], FeedbackStatusRepository, "Correct status repo instantiated" )
        self.assertEqual( obj.statusRepos[0].activity, self.activity, "Status repo instantiated with correct activity" )

        self.assertIsInstance(obj.activity_notifying_about, DiscussionReview, "Notifying about expected type of activity")
        self.assertEqual(obj.activity_notifying_about, self.unit.discussion_review, "Expected activity")

        self.assertIsInstance( obj.activity_for_review_pairings, DiscussionReview,
                               "Review pairings based on expected activity type: Discussion review" )
        self.assertEqual( obj.activity_for_review_pairings, self.unit.discussion_review, "Expected activity" )

        self.assertIsInstance( obj.activity, DiscussionReview,
                               "Working on results from expected  activity type" )
        self.assertEqual( obj.activity, self.unit.discussion_review, "Expected activity" )

    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.WorkRepositoryLoaderFactory' )
    def test_run_first_time_for_activity( self, workLoaderMock, studentRepoMock, messengerMock ):
        # even though this will be the result of the filtering
        # we need to limit here since the mock doesn't have the property
        self.workRepo.submitter_ids = self.student_ids
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )

        # setup students
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )
        # setup review pairings
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.students,
                                                                        obj.dao.session )
        obj.statusRepos = [create_autospec( FeedbackStatusRepository, previously_sent_students=[ ] )]

        # call
        obj.run()

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        # Every student should've been messaged
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.student_ids ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]

        # Check that all messages have the correct subject
        for sid, subj, body in messenger_args:
            expect_subj = FeedbackFromDiscussionReviewMessenger.email_subject_templ.format( self.unit.unit_number )
            self.assertEqual( expect_subj, subj, "Correct subject line" )

        # Status repo calls on messenger
        obj.messenger.status_repositories[0].record.assert_called()
        call_list = obj.messenger.status_repositories[0].record.call_args_list
        status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        self.assertEqual( len( self.student_ids ), len( call_list ),
                          "Status repo record called expected number of times" )
        for sid in self.student_ids:
            self.assertIn( sid, status_args, "StatusRepo.record called on all students" )

        # student repo calls on messenger
        for sid in self.student_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # This is the review which the assessor has submitted
            author_text = self.workRepo.get_formatted_work_by( record.assessor_id )
            # which we are sending to the person who created the posts
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( author_text )
            self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )

    # @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.FeedbackStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.WorkRepositoryLoaderFactory' )
    def test_run_previously_sent( self, workLoaderMock, studentRepoMock, messengerMock ):
        authors_who_have_not_been_notified = self.new_students_ids
        authors_who_have_been_previously_notified = self.preexisting_student_ids

        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )

        # setup students
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )
        # setup review pairings
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.students,
                                                                        obj.dao.session )
        reviewers_with_notified_authors = [r.assessor_id for r in preexisting_pairings if r.assessee_id in authors_who_have_been_previously_notified]
        reviewers_without_notified_authors = [s for s in self.student_ids if s not in reviewers_with_notified_authors]
        obj.statusRepos = [create_autospec( FeedbackStatusRepository, reviewers_with_notified_authors=reviewers_with_notified_authors )]

        # This will be handled by the _filter_notified step. But since
        # we are using a dummy work repo, calling remove_student_records won't do anything
        # Thus we need to pretend that it did its job. (We've tested it elsewhere)
        self.workRepo.submitter_ids = reviewers_without_notified_authors
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # call
        obj.run()

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( authors_who_have_not_been_notified ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]

        # Check that all messages have the correct subject
        for sid, subj, body in messenger_args:
            expect_subj = FeedbackFromDiscussionReviewMessenger.email_subject_templ.format( self.unit.unit_number )
            self.assertEqual( expect_subj, subj, "Correct subject line" )

        # Status repo calls on messenger which
        # record the message having been sent to authors
        obj.messenger.status_repositories[0].record.assert_called()
        call_list = obj.messenger.status_repositories[0].record.call_args_list
        status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        self.assertEqual( len( authors_who_have_not_been_notified ), len( call_list ),
                          "Status repo record called expected number of times" )
        for sid in authors_who_have_not_been_notified:
            self.assertIn( sid, status_args,
                           "StatusRepo.record called on authors who have now received their feedback (i.e., from newly submitted reviewers)" )

        # student repo calls on messenger
        for sid in authors_who_have_not_been_notified:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # This is the review which the assessor has submitted
            author_text = self.workRepo.get_formatted_work_by( record.assessor_id )
            # which we are sending to the person who created the posts
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( author_text )
            self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )

    #
    # @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StatusRepository' )
    #     @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    #     @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    #     @patch( 'CanvasHacks.SkaaSteps.SendDiscussionReviewToPoster.WorkRepositoryLoaderFactory' )
    #     def test_run( self, workLoaderMock, studentRepoMock, messengerMock, feedbackStatusRepoMock ):
    #         self.workRepo.submitter_ids = self.new_students_ids
    #
    #         workLoaderMock.make = MagicMock( return_value=self.workRepo )
    #
    #         # prepare student repo
    #         students = { s.student_id: s for s in self.students }
    #
    #         def se( sid ):
    #             return students.get( sid )
    #
    #         # studentRepoMock.get_student = MagicMock(side_effect=se)
    #         # studentRepoMock.download = MagicMock( return_value=self.students )
    #
    #         # call
    #         obj = SendDiscussionReviewToPoster( course=self.course, unit=self.unit, is_test=True, send=True )
    #         # obj.studentRepo = MagicMock()
    #         obj.studentRepo.get_student = MagicMock( side_effect=se )
    #         obj.studentRepo.download = MagicMock( return_value=self.students )
    #         self.preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.students,
    #                                                                         obj.dao.session )
    #
    #         obj.run()
    #
    #         # check
    #
    #         # Check that new assignments don't involve students who haven't submitted
    #         for rec in obj.associations:
    #             self.assertNotIn( rec.assessor_id, self.preexisting_student_ids,
    #                               "Newly assigned assessor not among previously assigned students" )
    #             self.assertNotIn( rec.assessee_id, self.preexisting_student_ids,
    #                               "Newly assigned assessee not among previously assigned students" )
    #
    #         # Check whether each new student has been assigned
    #         new_assessor_ids = [ r.assessor_id for r in obj.associations ]
    #         new_assessee_ids = [ r.assessee_id for r in obj.associations ]
    #         self.assertEqual( len( set( new_assessor_ids ) ), len( self.new_students_ids ),
    #                           "No duplicate assessor assignments" )
    #         self.assertEqual( len( set( new_assessee_ids ) ), len( self.new_students_ids ),
    #                           "No duplicate assessee assignments" )
    #         for sid in self.new_students_ids:
    #             self.assertIn( sid, new_assessee_ids, "Student is an assessee" )
    #             self.assertIn( sid, new_assessor_ids, "Student is an assessor" )
    #
    #         # ================== Events on Messenger
    #         # Check that mocked objects were called with expected data
    #         messengerMock.assert_called()
    #         self.assertEqual( messengerMock.call_count, len( self.new_students_ids ),
    #                           "Send method called expected number of times" )
    #         messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
    #                            messengerMock.call_args_list ]
    #         # print(args)
    #
    #         # Check that all messages have the correct subject
    #         for sid, subj, body in messenger_args:
    #             self.assertEqual( FeedbackFromDiscussionReviewMessenger.email_subject, subj, "Correct subject line" )
    #
    #         # Status repo calls on messenger
    #         obj.messenger.status_repositories.record.assert_called()
    #         call_list = obj.messenger.status_repositories.record.call_args_list
    #
    #         # obj.messenger.status_repositories.record_invited.assert_called()
    #         # call_list = obj.messenger.status_repositories.record_invited.call_args_list
    #         status_args = [ c[ 0 ][ 0 ] for c in call_list ]
    #         self.assertEqual( len( self.new_students ), len( call_list ),
    #                           "Status repo record_invited called expected number of times" )
    #         for sid in self.new_students_ids:
    #             self.assertIn( sid, status_args, "StatusRepo.record_invited called on all students" )
    #
    #         # student repo calls on messenger
    #         for sid in self.new_students_ids:
    #             obj.messenger.student_repository.get_student.assert_any_call( sid )
    #
    #         # Check the content sent
    #         for record in obj.associations:
    #             # print(record.assessee_id)
    #             author_text = self.workRepo.get_formatted_work_by( record.assessee_id )
    #             # see if sent to assessor
    #             sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
    #             rx = r'{}'.format( author_text )
    #             self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )




"""
Created by adam on 3/15/20
"""
__author__ = 'adam'

from unittest import TestCase

from CanvasHacks.Text.process import WordbagMaker
from CanvasHacks.Text.stats import WordCount
from TestingBase import TestingBase

if __name__ == '__main__':
    pass


class TestWordCount( TestingBase ):
    def setUp(self) -> None:
        self.config_for_test()

    def test_initialized_correctly( self ):
        obj = WordCount(count_stopwords=False)
        self.assertIsInstance(obj.bagmaker, WordbagMaker, "instantiated correct helper")
        self.assertFalse(obj.bagmaker.count_stopwords, 'param set on bagmaker')


    def test_analyze( self ):
        obj = WordCount(count_stopwords=True)

        expected = self.fake.random.randint(1, 1000)
        content = " ".join([ self.fake.word() for _ in range(0, expected)])

        result = obj.analyze(content)

        self.assertEqual(expected, result, 'expected count returned')




"""
Created by adam on 2/10/20
"""
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions
# from CanvasHacks.Loaders.quiz import drop_columns_from_frame
from CanvasHacks.Processors.quiz import drop_columns_from_frame
from CanvasHacks.TimeTools import getDateForMakingFileName

__author__ = 'adam'

import json
import time

import pandas as pd

from CanvasHacks.Files.FileTools import makeDataFileIterator, create_folder, get_newest_file
from CanvasHacks.Api.RequestTools import *


# -------------------------- Getting the report file from server
def report_url_gen( starting_report_url ):
    """
    Returns the original url followed by copies with the report id incremented by 1
    starting_report_url: 'https://canvas.csun.edu/api/v1/courses/85210/quizzes/165098/reports/92640'
    """
    r = starting_report_url.split( '/' )
    url_stub = "/".join( r[ :-1 ] )
    rid = int( r[ -1: ][ 0 ] )

    while True:
        yield "{}/{}?include[]=progress&include[]=file".format( url_stub, rid )
        rid += 1


def parse_download_url( response ):
    """Takess a requests.response from get_report_download_url
    and tries to parse out the file download url.
    todo: set this up to poll for progress
    """
    c = json.loads( response.content )

    try:
        download_url = c[ 'file' ][ 'url' ]
        print( download_url )
        return download_url

    except KeyError:
        # todo Poll the progress url until have download url
        #         progress_url = c['progress_url']
        # print( "Error: No download url provided ") #, c )
        pass

def get_report_download_url( report_url ):
    """Uses the report url to get the url from which the csv file can be downloaded"""
    #     data = {"include": ["progress", "file"]}
    #     data = {"include": "file"}
    url = "{}?include[]=progress&include[]=file".format( report_url )
    data = { }

    try:
        # Make the request
        # We can't use the usual function since we don't want response.json()
        response = requests.get( report_url, headers=make_request_header() )

        # print( 'response was ', response.status_code )
        response.raise_for_status()
        # todo or put check the progress url until workflow_state="completed"
        #         if response.progress.workflow_state != 'completed':

        # Continuing on since was successful
        return parse_download_url( response )

    except HTTPError as http_err:
        pass
        # print( f'HTTP error occurred: {http_err}' )  # Python 3.6

    except Exception as err:
        pass
        # print( f'Other error occurred: {err}' )  # Python 3.6


def download_report( download_url, save_file_path=None ):
    """Once we have the url to download the csv file, this
    will handle the download and load the data into a pandas
    dataframe
    """
    # Request the results file for unit
    # Open it with pandas
    frame = pd.read_csv( download_url )

    if save_file_path:
        # Temporary location to store the downloaded file
        frame.to_csv( save_file_path )

    return frame


def retrieve_quiz_data( quiz, rest_timeout=60, max_id_attempts=20, **kwargs ):
    """Returns a dataframe of the student report
    ACTUALLY WORKING VERSION!
    WILL ONLY WORK IF THE GENERATE REPORT BUTTON HAS BEEN MANUALLY CLICKED FIRST
    Returns:
        DataFrame

    """
    # This should request that the reports be generated
    reports = quiz.get_all_quiz_reports()

    print( "Resting for {} seconds while waiting for canvas to generate report".format( rest_timeout ) )
    time.sleep( rest_timeout )

    # The first report should be the student_analysis report. However,
    # for some reason, canvas will not return the file info with the
    # report we just created. However, it seems that the creation process
    # also created a report with a different id, which will contain the file info
    # At least in testing, this report id was 2 more than the student report
    # (the item_analysis report id was the student_analysis id + 1)
    # But will use a generator to cover more possibilities
    url_gen = report_url_gen( reports[ 0 ].url )

    for _ in range( 0, max_id_attempts ):
        # first value out of generator will be the original url
        url = next( url_gen )
        # print( "trying: ", url )

        # We request the hidden report object which will have the url for downloading
        # and parse out the url
        download_url = get_report_download_url( url )

        # We load the url and parse into a dataframe
        if download_url:
            return download_report( download_url )


def save_downloaded_report( activity, frame, SECTION=False ):
    """If we've downloaded the report programmatically, this
    saves it to the expected location
    """
    # save to file
    create_folder( activity.folder_path )
    fp = "{}/{}-{}-student-work.csv".format( activity.folder_path, getDateForMakingFileName(), activity.safe_name )
    # try:
    #     # if there's a particular section
    #     fp = "{}/{}-student-work.csv".format( activity_inviting_to_complete.folder_path, SECTION )
    # except NameError:
    #     fp = "{}/{}-{}-student-work.csv".format( activity_inviting_to_complete.folder_path, getDateForMakingFileName(), activity_inviting_to_complete.safe_name )

    try:
        frame.to_csv( fp )
    except Exception as e:
        print( "Error saving student work to file ", e )


# -------------------------- Handle stuff stored in a file on disk
def get_newest_data( activity ):
    # get data from newest file
    fiter = makeDataFileIterator( activity.folder_path )
    report_frames = [ ]
    try:
        while True:
            f = next( fiter )
            # print( "loading: ", f )
            frame = pd.read_csv( f )
            frame.submitted = pd.to_datetime( frame.submitted )
            if 'student_id' not in frame.index:
                frame.rename( { 'id': 'student_id' }, axis=1, inplace=True )
            # this makes it freak out for some reason
            #         frame.set_index('student_id', inplace=True)
            report_frames.append( frame )
    except StopIteration:
        return sort_frames_by_age( report_frames )[ 0 ]


def sort_frames_by_age( quiz_report_frames ):
    """Sorts the list of frames so that the newest
    is first, the oldest is last"""
    # assumes submitted is a datetime
    quiz_report_frames.sort( key=lambda x: x.submitted.max(), reverse=True )
    return quiz_report_frames


def get_whats_new( quiz_report_frames ):
    """Given a list of quiz report dataframes, returns a dataframe
    of the things that are new since the last time the data was checked
    """
    if len( quiz_report_frames ) == 1:
        return quiz_report_frames[ 0 ]
    sort_frames_by_age( quiz_report_frames )
    latest = quiz_report_frames[ 0 ]
    previous = quiz_report_frames[ 1 ]
    # Anything which happened after the last check of the data
    return latest[ latest.submitted > previous.submitted.max() ]


def load_new( activity ):
    """Returns a dataframe containing everything that is new in the most
    recent file since the next most recent file.

    NB, this is not equivalent to all the records which have not been
    acted upon since it is possible that no action was taken on the
    records in the next most recent file."""
    fiter = makeDataFileIterator( activity.folder_path )
    report_frames = [ ]
    try:
        while True:
            f = next( fiter )
            print( "loading: ", f )
            frame = pd.read_csv( f )
            frame.submitted = pd.to_datetime( frame.submitted )
            if 'student_id' not in frame.index:
                frame.rename( { 'id': 'student_id' }, axis=1, inplace=True )
            # this makes it freak out for some reason
            #         frame.set_index('student_id', inplace=True)
            report_frames.append( frame )
    except StopIteration:
        newstuff = get_whats_new( report_frames )
        print( "{} new records since previous file".format( len( newstuff ) ) )
        if len( newstuff ) == 0:
            raise NoNewSubmissions

        return newstuff


def load_activity_from_newest_file(activity):
    """
    Uses the creation time of the files in the activity's folder
    to determine which is the newest and loads data from it
    :param activity:
    :return:
    """
    fpath = get_newest_file(activity.folder_path)
    frame = pd.read_csv(fpath)
    frame.submitted = pd.to_datetime( frame.submitted )
    if 'student_id' not in frame.index:
        frame.rename( { 'id': 'student_id' }, axis=1, inplace=True )
    return frame


def load_activity_data_from_files( activity, course ):
    """Get complete set of data for activity_inviting_to_complete from a bunch of
    potentially inconsistent files.
    Created in CAN-41
    Loads the data into a repository object of the appropriate type
    """
    fiter = makeDataFileIterator( activity.folder_path )
    frames = [ ]
    try:
        while True:
            f = next( fiter )
            print( "loading: ", f )
            f = pd.read_csv( f )
            if 'student_id' not in f.columns:
                f.rename( { 'id': 'student_id' }, axis=1, inplace=True )
            frames.append( f )
    except StopIteration:
        print( "Loaded data from {} files".format( len( frames ) ) )

        data = pd.concat( frames, sort=True )
        print( "Loaded {} rows from all files in folder".format( len( f ) ) )

        if 'score_x' in data.columns:
            # this shouldn't be necessary in future.
            # Only when were saving sheet plus submissions
            def fix_score( row ):
                return row[ [ 'score', 'score_x', 'score_y' ] ].mean()

            data.score = data.apply( lambda x: fix_score( x ), axis=1 )
            data.drop( [ 'score_x', 'score_y' ], axis=1, inplace=True )

        # todo this stuff that's been commented out is probably essential to something (had to comment out in CAN-44 to prevent circular imports)
        # repo = WorkRepositoryFactory.make( activity_inviting_to_complete, course )
        # repo.set_question_columns( data )
        drop_columns_from_frame( data )

        data.drop_duplicates( inplace=True )
        # repo.data = data
        # print( "{} rows loaded to repo.data after processing".format( len( repo.data ) ) )
        # return repo
        return data


if __name__ == '__main__':
    pass





"""
Created by adam on 3/2/20
"""
__author__ = 'adam'



from CanvasHacks.Loaders.interfaces import ILoader

__author__ = 'adam'

from CanvasHacks.Repositories.discussions import DiscussionRepository


class DiscussionDownloadLoader( ILoader ):
    """Loads all records for quiz"""

    @classmethod
    def load( cls, activity, course, save=True, **kwargs ):
        """
        Handles loading data for a non-quiz unit
        :param activity:
        :param course:
        :param kwargs:
        :return: DataFrame
        """
        print( "Downloading posts for {}", activity.name )
        discussionRepo = DiscussionRepository( activity, course )
        discussionRepo.download()

        return discussionRepo

        #todo re-enable saving

        # We need the canvas.api.unit object
        # assignment = DiscussionDownloadLoader.get_assignment( course, activity )
        # subRepo = AssignmentSubmissionRepository( assignment )
        #
        # # parse out already graded submissions
        # # subRepo.data =[j for j in subRepo.data if j[0].grade != 'complete']
        # if save:
        #     # Want to have all the reports be formatted the same
        #     # regardless of whether we manually or programmatically
        #     # downloaded them. Thus we save before doing anything to them.
        #     save_downloaded_report( activity, subRepo.frame )

        # return subRepo.frame



class DiscussionFileLoader( ILoader ):
    """Loads all records for discussion from file"""

    @classmethod
    def load( cls, activity, course, save=True, **kwargs ):
        """
        Handles loading data for a non-quiz unit
        :param activity:
        :param course:
        :param kwargs:
        :return: DataFrame
        """
        # todo

        # print( "Downloading posts for {}", activity.name )
        # discussionRepo = DiscussionRepository( course, activity.topic_id )
        # discussionRepo.download()

        #todo re-enable saving

        # We need the canvas.api.unit object
        # assignment = DiscussionDownloadLoader.get_assignment( course, activity )
        # subRepo = AssignmentSubmissionRepository( assignment )
        #
        # # parse out already graded submissions
        # # subRepo.data =[j for j in subRepo.data if j[0].grade != 'complete']
        # if save:
        #     # Want to have all the reports be formatted the same
        #     # regardless of whether we manually or programmatically
        #     # downloaded them. Thus we save before doing anything to them.
        #     save_downloaded_report( activity, subRepo.frame )

        # return subRepo.frame







if __name__ == '__main__':
    pass



# todo This should be an end-to-end integration test which actually hits the server

# todo Should check the case where no student is available to review

# """
# Assigns a randomly selected reviewer and sends them the work
#
# """
# import unittest
#
# from unittest.mock import MagicMock, patch
#
# from CanvasHacks.PeerReviewed.Definitions import Unit, Review
# from tests.factories.PeerReviewedFactories import activity_data_factory, unit_factory
# from tests.factories.ModelFactories import student_factory
# import CanvasHacks.globals
# CanvasHacks.globals.use_api = False
#
# from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
# from tests.TestingBase import TestingBase
#
# from CanvasHacks.SkaaSteps.SendInitialWorkToReviewer import SendInitialWorkToReviewer
#
#
# class TestOnTimeSubmissions(TestingBase ):
#     """Checks that works properly on first run after
#     deadline on work that has been submitted
#     """
#
#     def setUp(self):
#         self.config_for_test()
#         self.dao = SqliteDAO()
#
#         self.course = MagicMock()
#         # review = Review(**activity_data_factory())
#         self.unit = unit_factory()
#         # self.unit.components.append(review)
#
#     @patch('CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository')
#     @patch('CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.PeerReviewInvitationMessenger')
#     @patch('CanvasHacks.SkaaSteps.ISkaaSteps.AssociationRepository')
#     @patch('CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory')
#     def test_sends_correctly( self, workLoaderMock, assocRepoMock, messengerMock, studentRepoMock ):
#         """Check that each student receives the expected message
#         containing the correct student's submission
#         """
#         students = [student_factory(), student_factory()]
#         submitter_ids = [s.student_id for s in students]
#         workRepo = MagicMock()
#         workRepo.submitter_ids = submitter_ids
#         workLoaderMock.make = MagicMock(return_value=workRepo)
#
#         studentRepoMock.download = MagicMock(return_value=students)
#         send = True
#         obj = SendInitialWorkToReviewer(course=self.course, unit=self.unit, is_test=True, send=True)
#
#         # call
#         obj.run()
#
#         # check
#         workLoaderMock.make.assert_called()
#         workLoaderMock.make.assert_called_with(self.unit.initial_work, self.course, True)
#
#         obj.studentRepo.download.assert_called()
#
#         obj.associationRepo.assign_reviewers.assert_called()
#         obj.associationRepo.assign_reviewers.assert_called_with(submitter_ids)
#
#         obj.messenger.notify.assert_called()
#         obj.messenger.notify.assert_called_with(obj.associationRepo.data, send)
#
#
#
#     #
#     # def test_assignments_stored_correctly( self ):
#     #     """Check that the students who have submitted the unit
#     #     (and none of the non-submitteers) have been paired up
#     #     in the expected manner, and that the pairings have been stored
#     #     """
#     #     self.fail()
#
#
# class TestLateSubmissions( TestingBase ):
#     """Checks that works properly on subsequent runs after the
#     initial assignments have been sent out
#
#     """
#
#     def test_sends_correctly( self ):
#         """Check that each student receives the expected message
#         containing the correct student's submission
#         """
#         self.fail()
#
#     def test_review_pairings_made_correctly( self ):
#         """Check that the students who have submitted the unit
#         (and none of the non-submitteers) have been paired up
#         in the expected manner, and that the pairings have been stored
#         """
#         self.fail()
#
#     def test_original_review_pairings_are_unaffected( self ):
#         """Check that the review assignments which were created
#         on the earlier run(s) have not been altered when the
#         new review assignments were made
#         """
#         self.fail()
#
#
# if __name__ == '__main__':
#     unittest.main()



"""
Created by adam on 2/25/20
"""
from logbook import Logger

__author__ = 'adam'

if __name__ == '__main__':
    pass


class ILogger( object ):

    def __init__( self, **kwargs ):
        self.logger = Logger( self.name )

    def _process_kwargs( self, kwargs ):
        for key, value in kwargs.items():
            setattr( self, key, value )

    def log( self, msg ):
        """Records record at notice level"""
        self.logger.notice( msg )

    def log_error( self, msg ):
        """Records record at error level"""
        self.logger.error( msg )

    def write( self, msg ):
        """Writes to a file"""
        raise NotImplementedError



"""
Created by adam on 2/22/20
"""
__author__ = 'adam'

import CanvasHacks.testglobals
from CanvasHacks.Repositories.status import StatusRepository, FeedbackStatusRepository, InvitationStatusRepository

CanvasHacks.testglobals.use_api = False
import unittest
from unittest.mock import MagicMock, patch
from tests.TestingBase import TestingBase

from tests.factories.ModelFactories import student_factory
from tests.factories.PeerReviewedFactories import unit_factory
from tests.factories.RepositoryMocks import ContentRepositoryMock

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO

from CanvasHacks.SkaaSteps.SendReviewToReviewee import SendReviewToReviewee
import CanvasHacks.environment as env
from faker import Faker

fake = Faker()


class TestCallsAllExpected( TestingBase ):
    """Makes sure that everything gets called with expected values.
       Not super diagnostic since many of the mocked calls are
       where we'd actually expect failure. Still, useful for catching
       problems when update code etc"""

    def setUp( self ):
        self.config_for_test()
        self.dao = SqliteDAO()

        self.course = MagicMock()
        self.unit = unit_factory()

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.MetareviewInvitationMessenger' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.AssociationRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.WorkRepositoryLoaderFactory' )
    def test_run( self, workLoaderMock, assocRepoMock, messengerMock, studentRepoMock ):
        """Check that each student receives the expected message
        containing the correct student's submission
        """
        students = [ student_factory(), student_factory() ]
        submitter_ids = [ s.student_id for s in students ]
        workRepo = MagicMock()
        workRepo.submitter_ids = submitter_ids
        workLoaderMock.make = MagicMock( return_value=workRepo )

        studentRepoMock.download = MagicMock( return_value=students )
        send = True
        obj = SendReviewToReviewee( course=self.course, unit=self.unit, is_test=True, send=True )
        # ra = ReviewAssociation(assessor_id=students[0], assessee_id=students[1])
        # obj.associationRepo.get_assessor_object = MagicMock(return_value=[ra])

        # call
        obj.run()

        # check
        workLoaderMock.make.assert_called()
        workLoaderMock.make.assert_called_with( self.unit.review, self.course) # only_new=False, rest_timeout=5 )

        obj.studentRepo.download.assert_called()

        obj.messenger.notify.assert_called()
        obj.messenger.notify.assert_called_with( obj.associations, send )


class TestUnitTests(TestingBase):
    """Unit tests with all dependencies mocked"""
    def setUp( self ):
        self.config_for_test()
        env.CONFIG.semester_name = "T30"
        self.unit = unit_factory()
        self.course = MagicMock()
        self.activity_id = self.unit.review.id
        # self.dao = SqliteDAO()
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    def test_instantiates_correct_status_repos( self ):
        """The sending of metareview results requires a
        special status repository"""
        obj = SendReviewToReviewee( course=self.course, unit=self.unit, is_test=True, send=True )

        self.assertIsInstance(obj.statusRepos, list, "status repos is a list")
        self.assertTrue(len(obj.statusRepos) == 2)
        self.assertIsInstance(obj.invite_status_repo, InvitationStatusRepository, "Invite repo instantiated")
        self.assertIsInstance( obj.feedback_status_repo, FeedbackStatusRepository, "Correct status repo instantiated" )




class TestFunctionalTests( TestingBase ):
    """Checks that works properly on first run after
    deadline on work that has been submitted
    """

    def setUp( self ):
        self.config_for_test()
        env.CONFIG.semester_name = "T30"
        self.unit = unit_factory()
        self.course = MagicMock()
        self.activity_id = self.unit.review.id
        # self.dao = SqliteDAO()
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.InvitationStatusRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.FeedbackStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.WorkRepositoryLoaderFactory' )
    def test_run( self, workLoaderMock, studentRepoMock, messengerMock, statusRepoMock, inviteStatusRepoMock ):

        # Prepare fake work repo to give values to calling  objects
        workRepo = ContentRepositoryMock()
        workRepo.create_test_content( self.student_ids )
        workRepo.submitter_ids = self.student_ids
        workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # call
        obj = SendReviewToReviewee( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        obj.run()

        # check
        self.assertEqual( len( obj.associations ), len( self.students ), "Correct number of students notified" )

        # Check that filtered previously notified
        workRepo.remove_student_records.assert_called()

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.students ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]
        # Check that all messages have the correct subject
        for sid, subj, body in messenger_args:
            self.assertEqual(self.unit.metareview.email_subject, subj, "Correct subject line")

        # Status repo calls on messenger
        for repo in obj.messenger.status_repositories:
            repo.record.assert_called()
            # obj.messenger.status_repositories.record_invited.assert_called()
            call_list = repo.record.call_args_list
            # call_list = obj.messenger.status_repositories.record_invited.call_args_list
            status_args = [ c[ 0 ][ 0 ] for c in call_list ]
            self.assertEqual( len( self.students ), len( call_list ),
                              "Status repo record called expected number of times" )
            for sid in self.student_ids:
                self.assertIn( sid, status_args, "Record called on both repos for all students" )

        # student repo calls on messenger
        for sid in self.student_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            reviewer_text = workRepo.get_formatted_work_by( record.assessor_id )
            # see if sent to assessee
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( reviewer_text )
            self.assertRegex( sent_text, rx, "Reviewer's work in message sent to author" )

    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.WorkRepositoryLoaderFactory' )
    def test_run_some_already_notified( self, workLoaderMock, studentRepoMock, messengerMock ):
        # Set up status repo with some students already notified
        num_previously_sent = fake.random.randint( 1, len( self.students ) - 1 )
        previously_sent = self.student_ids[ : num_previously_sent ]
        # previously_sent = fake.random.choices(self.student_ids, k=num_previously_sent)
        to_notify = [ s for s in self.student_ids if s not in previously_sent ]
        num_to_notify = len( to_notify )

        # Prepare fake work repo to give values to calling  objects
        workRepo = ContentRepositoryMock()
        workRepo.create_test_content( self.student_ids )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendReviewToReviewee( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        reviewers_without_prev_notified_authors = [s.assessor_id for s in self.preexisting_pairings if s.assessee_id not in previously_sent]

        # setting this since filter step doesn't happen on dummy
        workRepo.submitter_ids = reviewers_without_prev_notified_authors
        # workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=workRepo )

        # Set up previous notifications
        for sid in previously_sent:
            obj.feedback_status_repo.record( sid )

        # call
        obj.run()

        # check
        self.assertEqual( num_to_notify, len( obj.associations ), "Correct number of students notified" )

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, num_to_notify, "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in messengerMock.call_args_list ]

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            reviewer_text = workRepo.get_formatted_work_by( record.assessor_id )
            # see if sent to assessee
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( reviewer_text )
            self.assertRegex( sent_text, rx, "Reviewer's work in message sent to author" )

    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendReviewToReviewee.WorkRepositoryLoaderFactory' )
    def test_run_some_not_turned_in( self, workLoaderMock, studentRepoMock, messengerMock ):
        # Set up status repo with some students already notified
        num_unsubmitted = fake.random.randint( 1, len( self.students ) - 1 )
        submitted = self.student_ids[ : num_unsubmitted ]
        # previously_sent = fake.random.choices(self.student_ids, k=num_previously_sent)
        to_notify = [ s for s in self.student_ids if s in submitted ]
        num_to_notify = len( to_notify )
        # feedbackStatusRepoMock.previously_sent_result = MagicMock( return_value=previously_sent )

        # Prepare fake work repo to give values to calling  objects
        workRepo = ContentRepositoryMock()
        workRepo.create_test_content( self.student_ids )
        workRepo.submitter_ids = to_notify
        workLoaderMock.make = MagicMock( return_value=workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendReviewToReviewee( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        # call
        obj.run()

        # check
        self.assertEqual( num_to_notify, len( obj.associations ), "Correct number of students notified" )

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, num_to_notify, "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in messengerMock.call_args_list ]

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            reviewer_text = workRepo.get_formatted_work_by( record.assessor_id )
            # see if sent to assessee
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessee_id ][ 0 ]
            rx = r'{}'.format( reviewer_text )
            self.assertRegex( sent_text, rx, "Reviewer's work in message sent to author" )


if __name__ == '__main__':
    unittest.main()



"""
Created by adam on 1/22/20
"""
import csv

from CanvasHacks.Logging.interfaces import ILogger
from CanvasHacks.Logging.templates import entry_separator, error_entry_separator

__author__ = 'adam'

if __name__ == '__main__':
    pass

import CanvasHacks.environment as env


class CsvLogger( ILogger ):
    """Parent for loggers which write to csv file"""

    def __init__( self, log_file_path, logger_name='csvlogger', **kwargs ):
        self.logger_name = logger_name
        self.log_file_path = log_file_path

        # self.initialize_logger()

    def write( self, record ):
        """Writes a row to a csv log file
        Record should be a list of values
        """
        with open( self.log_file_path, 'a' ) as csvfile:
            writer = csv.writer( csvfile )
            writer.writerow( record )


class TextLogger( ILogger ):
    """
    Parent class for loggers which write to a textfile
    """

    def __init__( self, log_file_path=None, logger_name='csvlogger', **kwargs ):
        self.logger_name = logger_name
        # self.log_file_path = log_file_path

        # self.initialize_logger()

    @classmethod
    def initialize_logger( cls, log_file_path=None ):
        if log_file_path:
            cls.log_file_path = log_file_path
        t = "TEST-" if env.CONFIG.is_test else ""
        cls.log_file_path = "{}/{}{}".format( env.LOG_FOLDER, t, cls.log_file_name )

    # else:
        #     # todo raise error if LOG_FOLDER is none to indicate we want streaming log
        #     cls.log_file_path = env.STUDENT_WORK_PROCESSING_LOGNAME.format(env.LOG_FOLDER)

    @classmethod
    def write( cls, stuff, is_error=False ):
        """Writes to text log"""
        try:
            cls._actually_write( stuff, is_error )
        except AttributeError:
            cls.initialize_logger()
            cls._actually_write( stuff, is_error )

    @classmethod
    def _actually_write( cls, stuff, is_error=False ):
        """Abstracted out to allow automatic initialization"""
        with open( cls.log_file_path, 'a' ) as f:
            if is_error:
                f.write( error_entry_separator() )
                f.write( stuff.__str__() )
            else:
                f.write( entry_separator() )
                f.write( stuff )

    #
    # def initialize_logger( self ):
    #     try:
    #         if self.logger is not None:
    #             pass
    #     except:
    #         self.logger = Logger()
    #         # self.logger = FileHandler(self.log_file)
    #         # self.logger.push_application() #Pushes handler onto stack of log handlers
    #
    # def write( self, stuff ):
    #     with open( self.log_file_path, 'a' ) as f:
    #         f.write(entry_separator())
    #         f.write( stuff )
    #         # f.close()


class StudentWorkLogger( TextLogger ):
    """
    Handles logging student work
    """
    log_file_name = env.STUDENT_WORK_PROCESSING_LOGNAME

    # t = "TEST-" if env.CONFIG.is_test else ""
    # log_file_path = "{}/{}{}".format( env.LOG_FOLDER, t, env.STUDENT_WORK_PROCESSING_LOGNAME )

# ---------------------------- Decorators





"""
Created by adam on 5/6/19
"""
__author__ = 'adam'

from CanvasHacks import environment as env
import pandas as pd
from CanvasHacks.Models.model import Model

class StoredDataFileMixin:
    """Standard methods for accessing data stored in files"""

    @property
    def safe_name( self ):
        """Returns a name that won't f up the file path"""
        return "".join([n for n in self.name if n != ':'])

    @property
    def folder_path(self):
        # ssafename = "".join([n for n in self.name if n != ':'])
        return "{}/{}-{}".format(env.ARCHIVE_FOLDER, self.course_id, self.safe_name)


class QuizDataMixin:
    """ Methods for handling data downloaded for quiz type assignments """
    #
    # @property
    # def safe_name( self ):
    #     """Returns a name that won't f up the file path"""
    #     return "".join([n for n in self.name if n != ':'])
    #
    # @property
    # def folder_path(self):
    #     # ssafename = "".join([n for n in self.name if n != ':'])
    #     return "{}/{}-{}".format(env.ARCHIVE_FOLDER, self.course_id, self.safe_name)

    def set_question_columns(self, results_frame):
        """Finds the question columns in a results frame
        and stores them in a list of tuples
        with the form (question id, string column name)
        """
        questions = self._detect_question_columns(results_frame.columns)
        self.question_columns = [(q.split(':')[0], q) for q in questions]
        # weird stuff we need to filter
        bad = ['Unnamed']
        self.question_columns = [(a[0], a[1] ) for a in self.question_columns if a[0].strip() not in bad ]

    def _detect_question_columns(self, columns):
        """Return a list of columns which contain a colon,
        those probably contain the question answers
        """
        return [c for c in columns if len(c.split(':')) > 1]

    @staticmethod
    def _check_date( date ):
        """Checks that a value is a pd.Timestamp
        if not, it tries to make it into one"""
        return date if isinstance( date, pd.Timestamp ) else pd.to_datetime( date )

    @property
    def keep_columns( self ):
        """Returns a list of report column names which
        should not be dropped from the frame"""
        cols = [c[1] for c in self.question_columns]
        cols += env.REPORT_KEEP_COLUMNS
        return cols


class QuizData(Model, QuizDataMixin):
    """Holds the data which defines the properties of a
    canvas quiz type unit
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # self.handle_kwargs(kwargs)
        # for k in kwargs.keys():
        #     self.__setattr__(k, kwargs[k])
        self.question_columns = []

    @property
    def course_id(self):
        for a in self.html_url.split('/'):
            try:
                return int(a)
            except:
                pass

    @property
    def due_date(self):
        return type( self )._check_date(self.due_at)

    @property
    def open_date( self ):
        return type( self )._check_date(self.open_at)

    @property
    def quarter_credit_date(self):
        return self._quarter_credit_date

    @quarter_credit_date.setter
    def quarter_credit_date(self, date):
        self._quarter_credit_date = pd.to_datetime(date)
    #
    # @property
    # def folder_path(self):
    #     return "{}/{}-{}".format(env.ARCHIVE_FOLDER, self.course_id, self.name)

    @property
    def lock_date(self):
        return type( self )._check_date(self.lock_at)

    @property
    def max_score(self):
        return self.points_possible

    @property
    def name(self):
        return self.title

    # def set_question_columns(self, results_frame):
    #     """Finds the question columns in a results frame
    #     and stores them in a list of tuples
    #     with the form (question id, string column name)
    #     """
    #     questions = self._detect_question_columns(results_frame.columns)
    #     self.question_columns = [(q.split(':')[0], q) for q in questions ]
    #
    # def _detect_question_columns(self, columns):
    #     """Return a list of columns which contain a colon,
    #     those probably contain the question answers
    #     """
    #     return [c for c in columns if len(c.split(':')) > 1]
    #
    # @staticmethod
    # def _check_date( date ):
    #     """Checks that a value is a pd.Timestamp
    #     if not, it tries to make it into one"""
    #     return date if isinstance( date, pd.Timestamp ) else pd.to_datetime( date )



if __name__ == '__main__':
    pass


"""
Handles logging sent messages to students
Created by adam on 2/25/20
"""
from CanvasHacks import environment as env
from CanvasHacks.Logging.Logging import TextLogger

__author__ = 'adam'

if __name__ == '__main__':
    pass


class MessageLogger( TextLogger ):
    """All outgoing messages logged with this"""
    log_file_name = env.MESSAGE_LOGNAME
    #
    # t = "TEST-" if env.CONFIG.is_test else ""
    # log_file_path = "{}/{}{}".format( env.LOG_FOLDER, t, env.MESSAGE_LOGNAME )



"""
Created by adam on 9/21/18
"""
__author__ = 'adam'

import string

import nltk
from nltk.corpus import stopwords

from CanvasHacks.Errors.grading import NonStringInContentField



class ITextProcessor:

    def process( self, content ):
        raise NotImplementedError


class WordbagMaker( ITextProcessor ):

    def __init__( self, count_stopwords=True, remove=[ ], **kwargs ):
        self._remove = remove
        self.count_stopwords = count_stopwords

    def process( self, content ):
        """
        Returns a list of words
        :param content:
        :return:
        """
        if not isinstance( content, str ):
            raise NonStringInContentField

        return [ word.lower() for sent in nltk.tokenize.sent_tokenize( content ) for word in
                 nltk.tokenize.word_tokenize( sent ) if word.lower() not in self.to_remove ]

    @property
    def to_remove( self ):
        """
        Strings that should be filtered out when tokenizing

        :return:
        """
        self._remove += [ '``', "''", "'s" ]
        self._remove += string.punctuation
        if not self.count_stopwords:
            self._remove += stopwords.words( 'english' )
        return set( self._remove )




# ------------ old

def make_wordbag( text, to_remove=[] ):
    if len(to_remove) == 0:
        to_remove = [ '``', "''", "'s" ]
        to_remove += string.punctuation
        to_remove += stopwords.words( 'english' )
        to_remove = set( to_remove )

    return [ word.lower() for sent in nltk.tokenize.sent_tokenize( text ) for word in
             nltk.tokenize.word_tokenize( sent ) if word.lower() not in to_remove ]

if __name__ == '__main__':
    pass



"""
Created by adam on 3/10/20
"""
__author__ = 'adam'

from unittest.mock import create_autospec, MagicMock

from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.GradingCorrections.penalities import IPenalizer
from CanvasHacks.GradingHandlers.review import ReviewGrader
from CanvasHacks.Repositories.interfaces import ISubmissionRepo
from TestingBase import TestingBase
from factories.RepositoryMocks import ContentRepositoryMock

if __name__ == '__main__':
    pass


class TestReviewGrader( TestingBase ):
    def setUp( self ):
        self.config_for_test()
        self.points_per_question = self.fake.random.randint( 1, 100 )

        self.create_new_and_preexisting_students()

        self.penalizer = create_autospec( IPenalizer )
        self.grade_method = create_autospec( IGradingMethod )

        self.activity = MagicMock( grade_method=self.grade_method, penalizer=self.penalizer )

        self.work_repo = ContentRepositoryMock( activity=self.activity, points_per_question=self.points_per_question )
        self.work_repo.create_quiz_repo_data( self.student_ids,
                                            submitted_at=self.fake.date_time_this_century(),
                                              make_dataframe=True )

        self.submission_repo = create_autospec( ISubmissionRepo )

        self.obj = ReviewGrader( self.work_repo, self.submission_repo )

    def test__get_score( self ):
        content = self.fake.paragraph()
        # call
        self.obj._get_score( content )
        # check
        self.grade_method.grade.assert_called()
        self.grade_method.grade.assert_called_with( content )

    def test__get_score_custom_on_empty( self ):
        # Custom on empty case
        self.grade_method.grade = MagicMock( return_value=None )
        content = self.fake.paragraph()
        on_empty = 'taco'

        # call
        result = self.obj._get_score( content, on_empty=on_empty )

        # check
        self.assertEqual( on_empty, result, "returns custom value when empty" )
        self.grade_method.grade.assert_called()
        self.grade_method.grade.assert_called_with( content )

    def test__grade_row_no_penalty( self ):
        fudge_points = 0
        self.penalizer.get_fudge_points = MagicMock( return_value=fudge_points )
        self.grade_method.grade = MagicMock( return_value=self.points_per_question )
        row = self.work_repo.data.iloc[ 0 ]

        # call
        result = self.obj._grade_row( row )

        # check things called with expected values
        self.grade_method.grade.assert_called()
        for qid, column_name in self.work_repo.question_columns:
            # grade method called on each question field in row
            self.grade_method.grade.assert_any_call( row[ column_name ] )

        # check fudge points
        result_fudge_points = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'fudge_points' ]
        self.assertEqual( result_fudge_points, 0, "zero fudge points returned" )

        # check question scores
        result_question_dict = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'questions' ]
        for qid, cname in self.work_repo.question_columns:
            self.assertIn( qid, result_question_dict, "expected dictionary key" )
            self.assertIn( 'score', result_question_dict[ qid ], "expected internal dict key" )
            self.assertEqual( result_question_dict[ qid ][ 'score' ], self.points_per_question,
                              "expected question score" )

    def test_grade_row_penalty( self ):
        fudge_points = -50
        self.penalizer.get_fudge_points = MagicMock( return_value=fudge_points )
        self.grade_method.grade = MagicMock( return_value=self.points_per_question )
        row = self.work_repo.data.iloc[ 0 ]

        # call
        result = self.obj._grade_row( row )

        # check things called with expected values
        self.grade_method.grade.assert_called()
        for qid, column_name in self.work_repo.question_columns:
            # grade method called on each question field in row
            self.grade_method.grade.assert_any_call( row[ column_name ] )

        # check fudge points
        result_fudge_points = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'fudge_points' ]
        self.assertEqual( result_fudge_points, fudge_points, "expected fudge points returned" )

        # check question scores
        result_question_dict = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'questions' ]
        for qid, cname in self.work_repo.question_columns:
            self.assertIn( qid, result_question_dict, "expected dictionary key" )
            self.assertIn( 'score', result_question_dict[ qid ], "expected internal dict key" )
            self.assertEqual( result_question_dict[ qid ][ 'score' ], self.points_per_question,
                              "expected question score" )

    def test__penalty_message( self ):
        self.skipTest( 'todo' )


class TestReviewGraderFunctionalTests( TestingBase ):

    def test_grade( self ):
        self.skipTest( 'todo' )
        # todo Mixed credit and no credit students
        # todo Different total scores
        # todo Different attempts



"""
Created by adam on 3/17/20
"""
__author__ = 'adam'

from CanvasHacks.Errors.grading import StudentUnableToComplete
from CanvasHacks.Errors.review_pairings import NoReviewPairingFound
from CanvasHacks.Repositories.mixins import ObjectHandlerMixin

if __name__ == '__main__':
    pass


class BlockedByOtherStudent(ObjectHandlerMixin):
    """
    For assignments where a student's ability to complete it depends
    on other students doing something, this determines whether a correction
    is needed, e.g., if we're on a metareview but the reviewer never turned it in
    """

    def __init__( self, submissions_repo, pairing_repo ):
        """
        :param submissions_repo: The submissions repository for the activity which our student's ability to complete depends on
        :param pairing_repo: Review pairings that create the dependency
        """
        self.submissions_repo = submissions_repo
        self.pairing_repo = pairing_repo

        # The field in the repo that we will check to determine
        # whether the student depended on did the work
        self.completion_field = 'workflow_state'
        # The value the completion field will have if completed
        self.complete_value = 'submitted'
        self.incomplete_value = 'unsubmitted'

    def analyze( self, student ):
        """
        Looks up whether the provided student is/was able to complete
        the assignment.

        Will raise exceptions to make it easier to bubble
        them up to various handlers (e.g., one which notifies the blocked
        and blocking student)

        :param student: Student object or int
        :return: boolean
        :raises StudentUnableToComplete, NoReviewPairingFound
        """
        student_id = self._handle_id(student)

        ra = self.pairing_repo.get_by_author(student_id)
        depends_on = ra.assessor_id

        if depends_on is None:
            # todo If this is true the student was never assigned a reviewer, what should be done?
            raise NoReviewPairingFound(student_id)
            # return False

        blocked = not self.submissions_repo.has_completed(student_id)

        # dependency_student_row = self.submissions_repo.get_student_work( student_id )
        #
        # if dependency_student_row[self.completion_field] == self.complete_value:
        #     return True
        if blocked:
            raise StudentUnableToComplete(blocked_student=student_id, blocking_student=depends_on)

        return False




"""
Created by adam on 2/28/20
"""
__author__ = 'adam'

from CanvasHacks.Logging.messages import MessageLogger
from CanvasHacks.Messaging.SendTools import ConversationMessageSender
from CanvasHacks.Models.student import get_first_name
from CanvasHacks.Definitions.unit import Unit

if __name__ == '__main__':
    pass


class SkaaMessenger:

    def __init__( self, unit: Unit, student_repository, content_repository,
                  status_repositories: list ):
        """
        :param unit:
        :param student_repository:
        :param content_repository:
        :param status_repositories: List of status repos to call once sent
        """
        self.unit = unit
        self.student_repository = student_repository
        self.content_repository = content_repository

        # Object responsible for actually sending message
        self.sender = ConversationMessageSender()

        # Objects in charge of storing change in status
        # after sent. Should be a list of InvitationStatusRepository
        # or FeedbackStatusRepository objects
        self.status_repositories = status_repositories

        self.logger = MessageLogger()

    @property
    def sent_count( self ):
        return self.sender.sent_count

    @property
    def send_errors( self ):
        return self.sender.errors

    def _make_message_data( self, receiving_student, content, other=None ):
        """
        Creates a dictionary with data to be passed to the
        method which actually sends the info to the receiving student
        """
        message = self._make_message_content( content, other, receiving_student )

        return {
            'student_id': receiving_student.id,
            'subject': self.activity_inviting_to_complete.email_subject,
            'body': message
        }

    def _make_message_content( self, content, other, receiving_student ):
        d = self._make_template_input( content, other, receiving_student )
        message = self.message_template.format( **d )
        # message = make_notice( d )
        return message

    def _make_template_input( self, content, other, receiving_student ):
        """Creates the dictionary that will be used to format the message template
        and create the message content
        This is abstracted out to make testing easier
        """

        d = {
            'intro': self.activity_inviting_to_complete.email_intro,

            'name': get_first_name( receiving_student ),

            # Formatted work for sending
            'responses': content,

            # Add any materials from me
            'other': other if other is not None else "",

            # Add code and link to do reviewing unit
            'review_assignment_name': self.activity_inviting_to_complete.name,
            'access_code_message': self._make_access_code_message(),
            'review_url': self.activity_inviting_to_complete.html_url,
            'due_date': self.activity_inviting_to_complete.string_due_date
        }
        return d

    def _make_access_code_message( self ):
        """Adds text with the access code for the next unit if a code
        exists otherwise returns an empty string
        """
        tmpl = "Here's the access code: {}"

        if self.activity_inviting_to_complete.access_code is not None:
            return tmpl.format( self.activity_inviting_to_complete.access_code )
        return ""

    def prepare_message( self, review_assignment, other=None ):
        """Creates the message data for sending specific to the unit"""
        raise NotImplementedError

    def notify( self, review_assignments, send=False, other=None ):
        """Given a list of review unit objects, sends the
        appropriate notification message to the correct person
        for the unit
        """
        messages = [ ]
        for rev in review_assignments:
            # print( rev )
            message_data = self.prepare_message( rev, other )
            # messages.append( message_data )

            if send:
                m = self.sender.send( **message_data )
                # todo Decide whether to keep the logging on the sender.send method or add the following here so all outgoing messages are written to file. NB, if uncomment this, will need to change to use to call class method
                # self.logger.write(m)

                messages.append( m )
                self.update_status( message_data )

            else:
                # For test runs
                messages.append( message_data )
                print( message_data )
        # Returns for testing / auditing
        return messages

    def update_status( self, message_data ):
        """
        Call the record method on all status repos
        :param message_data:
        :return:
        """
        if self.status_repositories is not None:
            # If was passed a single repo, make a list
            if not isinstance(self.status_repositories, list):
                self.status_repositories = [ self.status_repositories ]

            for status_repo in self.status_repositories:
                # Record status change (not to log)
                # records that the student has been invited or that they
                # have been sent feedback depending on the class
                status_repo.record( message_data[ 'student_id' ] )



"""
Base for all tools which add to or reduce
a total score based on some criteria

Created by adam on 3/16/20
"""
__author__ = 'adam'

from abc import ABC

if __name__ == '__main__':
    pass


class IGradeCorrection(ABC):
    """
    GradeCorrections return a percentage that the
    initial total score should be raised.

    # todo Decide whether there is value in distinguishing between penalizers and corrections since they just return a different signed percentage

    """

    def analyze( self, *args, **kwargs ):
        """
        Returns a signed float of the percentage
        the contribution the evaluated item makes to the
        overall score should be raised or lowered.

        if overall score = 4, comprising
            A = 2
            B = 1
            C = 1
        and this correction applies to A, the result 0.5
        means their overall score should be 3

        :param work:
        :return: signed float
        """
        raise NotImplementedError



"""
Created by adam on 2/28/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass


class ISender:
    """Parent of all classes which handle sending
    messages to students"""

    def send( self ):
        """
        Handles sending through whatever channel
        :return:
        :raises MessageSendError which wraps the server's error
        """
        raise NotImplementedError


"""
Created by adam on 3/29/20
"""
__author__ = 'adam'

from CanvasHacks import environment
from CanvasHacks.Definitions.journal import Journal
from CanvasHacks.Models.model import StoreMixin


# todo dev

class GradeJournalStep(StoreMixin):
    """
    Executes the entire grading process for journals
    """


    def __init__( self, activity=None, rest_timeout=5, no_late_penalty=True, upload_grades=True, **kwargs ):
        self.upload_grades = upload_grades
        self.wait = rest_timeout
        self.no_late_penalty = no_late_penalty
        if activity is None:
            # Use what's stored in environment
            selected_activities = environment.CONFIG.get_assignment_ids()
            # todo update to allow multiple selected activities
            activity_id = selected_activities[0]
            activity = environment.CONFIG.unit.get_activity_by_id(activity_id)

        self.activity = activity

        if self.no_late_penalty:
            self.activity.due_at = self.activity.lock_at

        self.graded = [ ]

        self.handle_kwargs( **kwargs )

    def _initialize( self ):
        """
        Handle instantiating and loading repositories
        :return:
        """
        self.workRepo = WorkRepositoryLoaderFactory.make( course=environment.CONFIG.course,
                                                          activity=self.activity,
                                                          rest_timeout=self.wait )


        self.assignment = environment.CONFIG.course.get_assignment( self.activity.id )

        self.subRepo = AssignmentSubmissionRepository( self.assignment )
        # shove the activity onto a sub repo so it will resemble
        # a quizrepo for the grader
        self.subRepo.activity = self.activity


        # Filter previously graded
        self.subRepo.data = [ s for s in self.subRepo.data if s.workflow_state != 'complete' ]
        self.workRepo.data = self.workRepo.data[ self.workRepo.data.workflow_state != 'complete' ].copy( deep=True )

        self.workRepo.data.reset_index( inplace=True )

    def run( self, **kwargs ):

        from CanvasHacks.Repositories.submissions import SubmissionRepository
        results = [ ]

        for a in environment.CONFIG.assignments:
            # canvas api object
            assignment = self.course.get_assignment( int( a[ 0 ] ) )
            # activity object to define the features
            journal = Journal( **assignment.attributes )
            # Download submissions
            subRepo = SubmissionRepository( assignment )
            # parse out already graded submissions
            subRepo.data = [ j for j in subRepo.data if j.grade != 'complete' ]

            # shove the activity onto a sub repo so it will resemble
            # a quizrepo for the grader
            subRepo.activity = journal
            # Initialize the package for results
            store = DataStoreNew( journal )
            # provisionally determine credit
            grader = JournalGrader( subRepo )
            store.results = grader.grade()
            #     store.results = GT.new_determine_journal_credit(journal, subRepo)
            #     if GRADING_LATE:
            #         store.results = [j for j in store.results if j[0].grade != 'complete']

            results.append( store )

    def _upload_step( self ):

        self._log_uploaded()

    def _log_uploaded( self ):
        """
        Prints or writes to file what happened during upload
        :return:
        """
        stem = "Grades uploaded for {} students "
        print( stem.format( self.uploaded ) )


if __name__ == '__main__':
    pass


"""
Created by adam on 1/18/20
"""
__author__ = 'adam'

from sqlalchemy import Column, Integer
# from sqlalchemy.ext.declarative import declarative_base

from CanvasHacks import environment as env
from CanvasHacks.Models.model import Model
from CanvasHacks.Models.student import Student

from CanvasHacks.DAOs.sqlite_dao import Base
# Base = declarative_base()


class ReviewAssociation( Base, Model ):
    __tablename__ = env.REVIEW_ASSOCIATIONS_TABLE_NAME
    id = Column( Integer, primary_key=True )
    activity_id = Column( Integer )
    assessor_id = Column( Integer )
    assessee_id = Column( Integer )

    def __repr__( self ):
        return "<ReviewAssociation(id={}, activity_id={}, assessor_id={}, assessee_id={}".format(self.id, self.activity_id, self.assessor_id, self.assessee_id)

    @property
    def is_self_assignment( self ):
        """This will be an invalid unit
        but we will leave it up to something checking this property
        to raise an error or determine what to do"""
        return self.assessee_id == self.assessor_id

    # @property
    # def assessee( self ):
    #     return Student(self.assessee_id)
    #
    # @property
    # def assessor( self ):
    #     return Student(self.assessor_id)



if __name__ == '__main__':
    pass



"""
Created by adam on 2/15/20
"""
from CanvasHacks.GradingHandlers.base import IGrader

__author__ = 'adam'

from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.GradingCorrections.penalities import IPenalizer


class DiscussionForumGrader( IGrader ):
    """
    Grades discussion forum posts
    """

    grade_method: IGradingMethod
    penalizer: IPenalizer

    def __init__( self, work_repo, num_posts_required=1,  **kwargs ):
        """

        :param work_repo:
        :param num_posts_required:
        :param kwargs:
        """
        self.num_posts_required = num_posts_required
        self.work_repo = work_repo

        super().__init__(**kwargs)

        self.corrections = self.activity.corrections

        self.penalizers = self.activity.penalizers

        self.grade_methods = self.activity.grade_methods
        # Todo Check that output of grading methods sums to 1?

        # self.grade_method = self.activity.grade_method
        # self.penalizer = self.activity.penalizer

        self.credit_per_post = 100 / self.num_posts_required

        # Internal store for in between grading steps
        self._raw = [
            # ( sid, percent credit)
        ]

        # Externally usable store
        self.graded = []

    def _calc_scores( self ):
        """Iterate through repo data and store individual
        student scores in _raw
        """
        for p in self.work_repo.data:
            # discussion repo data will look like:
            # [{'student_id', 'student_name', 'text'}]
            pct_credit = 0
            for method in self.grade_methods:
                # each method should return a float pct
                # all of which sum to 1
                pct_credit += method.grade( p['text'],
                                            on_credit=self.credit_per_post,
                                            on_no_credit=0)

            # pct_credit = self.grade_method.grade(p['text'], on_credit=self.credit_per_post, on_no_credit=0)
            # pct_credit = self.credit_per_post if receives_credit( p[ 'text' ] ) else 0
            self._raw.append( (p[ 'student_id' ], pct_credit) )

    def grade( self ):
        """Assigns a provisional grade for the discussion unit
        Will return as a list of tuples
        todo: Add logging of details of how grade assigned
        """
        self._calc_scores()

        self._prepare_results()

        return self.graded

    def _prepare_results( self ):
        """Takes the raw scores and prepares them for upload
        """
        # sum them up for each student
        sids = list( set( [ s[ 0 ] for s in self._raw ] ) )
        for sid in sids:
            t = sum( [ s[ 1 ] for s in self._raw if s[ 0 ] == sid ] )
            # todo Should this be rounded to ceiling?
            t = round(t)
            # Correct for more than required number
            t = 100 if t > 100 else t
            self.graded.append( (sid, t) )


#
# def assign_grades(discussion_repo, num_posts_required):
#     """Assigns a provisional grade for the discussion unit
#     """
#     credit_per_post = 100 / num_posts_required
#
#     grades = [
#         # ( sid, percent credit)
#     ]
#
#     for p in discussion_repo.data:
#         #     print(post, receives_credit(post))
#         #         credit_per_post
#         pct_credit = credit_per_post if receives_credit(p['text']) else 0
#         grades.append( (p['student_id'], pct_credit) )
#
#     # sum them up and put in list for upload
#     totals = []
#     sids = list(set([s[0] for s in grades]))
#     for sid in sids:
#         t = sum([s[1] for s in grades])
#         t = 100 if t > 100 else t
#         totals.append( ( sid,  t))
#     return totals
#

if __name__ == '__main__':
    pass


"""
Created by adam on 5/6/19
"""
from CanvasHacks.Definitions.base import BlockableActivity
from CanvasHacks.Definitions.skaa import MetaReview
from CanvasHacks.Errors.grading import NonStringInContentField, StudentUnableToComplete
from CanvasHacks.GradingAnalyzers.blocked import BlockedByOtherStudent
from CanvasHacks.GradingCorrections.penalities import IPenalizer
from CanvasHacks.GradingHandlers.base import IGrader
from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.Logging.penalties import PenaltyLogger
from CanvasHacks.Repositories.interfaces import ISubmissionRepo
from CanvasHacks.Repositories.quizzes import QuizRepository
import CanvasHacks.testglobals

__author__ = 'adam'

if __name__ == '__main__':
    pass


class QuizGrader( IGrader ):
    grade_method: IGradingMethod
    penalizer: IPenalizer

    def __init__( self, work_repo: QuizRepository, submission_repo: ISubmissionRepo, association_repo=None, **kwargs ):
        """

        :param work_repo:
        :param submission_repo:
        :param association_repo: Only needed if grading an assignment which can be blocked
        :param kwargs:
        """
        self.association_repo = association_repo
        self.blocked_checker = None
        self.work_repo = work_repo
        self.submission_repo = submission_repo
        super().__init__( **kwargs )

        self.corrections = self.activity.corrections

        self.penalizers = self.activity.penalizers

        self.grade_methods = self.activity.grade_methods
        # Todo Check that output of grading methods sums to 1?

        self.graded = [ ]

        if isinstance( self.work_repo.activity, BlockableActivity ) and self.association_repo is not None:
            # association repo better have been set
            # todo
            self.blocked_checker = BlockedByOtherStudent( self.submission_repo, self.association_repo )

    def _compute_question_score( self, content, **kwargs ):
        """
        Calls the grading method which calculates the points received for
        a given question.s
        :param content:
        :return: integer points, None, or on_empty
        """
        pct_possible = 0
        for method in self.grade_methods:
            # each method should return a float pct
            # all of which sum to 1
            pct_possible += method.grade( content )
        # pct_possible = self.grade_method.grade( content )

        # return pct_possible
        return self.work_repo.points_per_question * pct_possible

    def _compute_penalty_pct( self, row, **kwargs ):
        down = 0
        # get a negative float representing the pct the total score
        # should be adjusted down
        for penalizer in self.penalizers:
            down += penalizer.analyze( **row.to_dict(), **kwargs )
        return down

    def _compute_correction_pct( self, row, **kwargs ):
        up = 0
        for correction in self.corrections:
            # get a positive float representing the pct the total score
            # should be adjusted up
            up += correction.analyze( **row.to_dict(), **kwargs )
        return up

    def _compute_initial_total( self, row, **kwargs ):
        """
        Returns the questions dict to be sent and the total
        score
        :param row:
        :param kwargs:
        :return:
        """
        # used for computing penalty
        total_score = 0
        questions = { }
        # Grade on emptiness
        for qid, column_name in self.work_repo.question_columns:
            content = row[ column_name ]
            pts = self._compute_question_score( content, **kwargs )
            # Quiz type things require the scores to be uploaded
            # for each question.
            # Thus we grade each question and then penalize the
            # overall score with fudge points if necessary
            questions[ qid ] = { 'score': pts }
            total_score += pts

        return total_score, questions

    def _grade_row( self, row, **kwargs ):
        """Grades a row

        NB, requires that the non-graded attempts be
        filtered out before passing in

        todo test whether the problem with non-graded people getting zeros is caused here

        :param row: pd.DataFrame row
        :param kwargs:
        :return: Dictionary formatted for uploading
        """

        # for student
        total_score = 0

        # compute the initial total score of the activity
        # this is the job of the GradingMethod
        total_score, questions = self._compute_initial_total( row, **kwargs )

        try:
            if total_score == 0 and self.blocked_checker is not None:
                # This will raise an exception if they are blocked
                # and we won't return anything
                self.blocked_checker.analyze( row[ 'student_id' ] )

            # Compute penalty if needed
            # Will be 0 if not docking for lateness
            # Records of penalty will be stored on self.penalizer.penalized_records
            down = self._compute_penalty_pct( row, **kwargs )

            # Calculate things which could increase the score
            # this also may raise an exception if the student was
            # unable to complete it
            up = self._compute_correction_pct( row, **kwargs )

            adj = up + down

            fudge_points = total_score * adj

            out = self._make_graded_row_output( row, questions, fudge_points )

            return out

        except StudentUnableToComplete as e:
            # handle situation where they were blocked by e.g. a reviewer not turning
            # in their part

            # NB this is going to be tricky once relax deadlines
            # Should initially just log everything. But there will need to
            # be a switch for when they get full credit
            print( e )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise e


    def _make_graded_row_output( self, row, question_score_dict, fudge_points ):
        out = {
            'student_id': int( row[ 'student_id' ] ),
            'attempt': int( row[ 'attempt' ] ),
            'submission_id': int( row[ 'submission_id' ] ),
            'course_id': int( row[ 'course_id' ] ),
            'quiz_id': int( row[ 'quiz_id' ] ),
            'data': { }
        }

        out[ 'data' ][ "quiz_submissions" ] = [
            {
                "attempt": int( row[ 'attempt' ] ),
                "fudge_points": fudge_points,
                "questions": question_score_dict
            }
        ]

        return out

        # if self.using_percentage:
        #     pct = 100 * adj
        #
        # # using total points
        # score = total_score * adj

        #
        # # Grade on emptiness
        # for qid, column_name in self.work_repo.question_columns:
        #     content = row[ column_name ]
        #     pts = self._compute_question_score( content, **kwargs )
        #     # Quiz type things require the scores to be uploaded
        #     # for each question.
        #     # Thus we grade each question and then penalize the
        #     # overall score with fudge points if necessary
        #     questions[ qid ] = { 'score': pts }
        #     total_score += pts
        #
        #     # Compute penalty if needed
        #     # Will be 0 if not docking for lateness
        #     # Records of penalty will be stored on self.penalizer.penalized_records
        #     fudge_points = self.penalizer.get_fudge_points(row['submitted'], total_score, row)
        #
        # out = self._make_graded_row_output(row, questions, fudge_points)
        #
        # return out


    def grade( self, **kwargs ):
        """
        Grades all rows in work_repo.data
        todo: Add logging of details of how grade assigned
        :return: List of objects ready for upload
        """

        for i, row in self.work_repo.data.iterrows():
            try:
                self.graded.append( self._grade_row( row, **kwargs ) )
            except NonStringInContentField as e:
                print( e, row )

        # Print or log any penalties applied.
        # The penalizer object leaves this task up to us
        PenaltyLogger.log( self.penalizer )

        return self.graded

    # def _penalty_message( self, penalty, row ):
    #     """
    #     Handles printing or logging of penalties applied
    #
    #     # todo enable logging
    #
    #     :param penalty:
    #     :param row:
    #     :return:
    #     """
    #     stem = 'Student #{}: Submitted on {}; was due {}. Penalized {}'
    #     return stem.format( row[ 'student_id' ], row[ 'submitted' ], self.activity.due_at, penalty )
    #

    # def report_late_penalties( self ):
    #     # Report late penalties
    #     if len( self.penalizer.penalized_records ) > 0:
    #         for penalty_dict in self.penalizer.penalized_records:
    #             self._penalty_message( penalty_dict[ 'penalty' ], penalty_dict[ 'record' ] )



"""
Created by adam on 3/9/20
"""
__author__ = 'adam'

from CanvasHacks.GradingHandlers.journal import JournalGrader
from TestingBase import TestingBase

from unittest.mock import MagicMock


from unittest import TestCase


if __name__ == '__main__':
    pass


class TestJournalGrader( TestingBase ):

    def make_journal_submission( self, is_empty=False ):
        txt = "" if is_empty else self.fake.paragraph()
        return MagicMock(body=txt,
                         submitted_at=self.fake.date_time_this_century())

    def setUp( self ):
        self.config_for_test()
        self.create_new_and_preexisting_students()

        self.full_credit_score = 100

        self.penalizer = MagicMock()
        self.grade_method = MagicMock()
        self.grade_method.grade = MagicMock(return_value=self.full_credit_score)
        self.activity = MagicMock(grade_method=self.grade_method, penalizer=self.penalizer)

    def test_grade( self ):
        data = [self.make_journal_submission() for _ in self.students]
        self.penalizer.get_penalized_score = MagicMock(return_value=self.full_credit_score)
        work_repo = MagicMock( activity=self.activity, data=data )
        obj = JournalGrader(work_repo)

        # call
        obj.grade()

        # check
        self.assertEqual(len(self.students), len(obj.graded), "Expected number in output")

        for d in data:
            # Check that grade method received expected value
            self.grade_method.grade.assert_any_call(d.body)
            # Check that penalizer method received expected value
            # self.penalizer.get_penalized_score.assert_any_call((d.submitted_at, self.full_credit_score))

        for g in obj.graded:
            # Check put in output
            self.assertEqual(self.full_credit_score, g[1], "Correct score added")

    def test_grade_penalty( self ):
        data = [self.make_journal_submission() for _ in self.students]
        penalized_score = 21
        self.penalizer.get_penalized_score = MagicMock(return_value=penalized_score)
        work_repo = MagicMock( activity=self.activity, data=data )
        obj = JournalGrader(work_repo)

        # call
        obj.grade()

        # check
        self.assertEqual(len(self.students), len(obj.graded), "Expected number in output")

        for d in data:
            # Check that grade method received expected value
            self.grade_method.grade.assert_any_call(d.body)
            # Check that penalizer method received expected value
            # self.penalizer.get_penalized_score.assert_any_call((d.submitted_at, self.full_credit_score))

        for g in obj.graded:
            # Check put in output
            self.assertEqual(penalized_score, g[1], "Correct penalized score added")






"""
Created by adam on 2/25/20
"""
import datetime

import pandas as pd

from CanvasHacks import environment as env
from CanvasHacks.TimeTools import getDateForMakingFileName

__author__ = 'adam'

if __name__ == '__main__':
    pass


def make_review_audit_file(associationRepo, studentRepo, unit):
    """Stores all review assignments to a csv file
    for easier auditing
    """
    d = associationRepo.audit_frame(studentRepo)
    d['timestamp'] = datetime.datetime.now().isoformat()

    fp = "{}/{}-Unit{}-peer review assignments.csv".format(env.LOG_FOLDER, getDateForMakingFileName(), unit.unit_number )
    d.to_csv(fp)
    print("Audit file saved to {}".format(fp))

    #
    # # Make audit file
    # review_audit = []
    # for rev in associationRepo.get_associations(unit.review):
    #     print(rev.assessor_id, rev.assessee_id)
    #     assessor = studentRepo.get_student_record( rev.assessor_id )
    #     assessee = studentRepo.get_student_record( rev.assessee_id )
    #     print(assessor)
    #
    #     review_audit.append({
    #         'activity_inviting_to_complete' : unit.review.name,
    #         'assessor_name' : assessor.short_name,
    #         'assessor_sis_id': assessor.sis_user_id,
    #         'assessor_canvas_id': assessor.id,
    #         'assessee_name' : assessee.short_name,
    #         'assessee_sis_id': assessee.sis_user_id,
    #         'assessee_canvas_id': assessee.id,
    #         'timestamp': datetime.datetime.now().isoformat()
    #     })
    #
    # review_audit = pd.DataFrame(review_audit)




"""
Created by adam on 2/22/20
"""
__author__ = 'adam'

from unittest.mock import MagicMock, patch

from faker import Faker

from CanvasHacks.Messaging.templates import REVIEW_NOTICE_TEMPLATE
from CanvasHacks.Messaging.skaa import PeerReviewInvitationMessenger
from CanvasHacks.Repositories.students import StudentRepository
from tests.TestingBase import TestingBase
from tests.factories.ModelFactories import student_factory
from tests.factories.PeerReviewedFactories import unit_factory
from tests.factories.RepositoryMocks import ContentRepositoryMock

fake = Faker()


class TestStudentWorkForPeerReviewMessenger( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        # self.activity_data = activity_data_factory()
        self.activity = self.unit.review #InitialWork( **self.activity_data )

        # student recieiving the message
        self.author = student_factory()
        self.reviewer = student_factory()

        # This would be the content unit
        self.work = fake.text()

        self.studentRepo = StudentRepository()
        self.studentRepo.get_student = MagicMock( return_value=self.reviewer )
        self.contentRepo = ContentRepositoryMock()
        self.contentRepo.get_formatted_work_by = MagicMock( return_value=self.work )
        self.review_assign = MagicMock( assessor_id=self.reviewer.id, assessee_id=self.author.id )
        self.statusRepo = MagicMock() #create_autospec(StatusRepository)

    def test_prepare_message( self ):
        # student recieiving the message
        self.obj = PeerReviewInvitationMessenger( self.unit, self.studentRepo, self.contentRepo, self.statusRepo )

        # call
        message_data = self.obj.prepare_message( self.review_assign )

        # print(message_data)

        # check
        self.assertEqual( self.obj.message_template, REVIEW_NOTICE_TEMPLATE, "Working off expected template" )
        self.assertEqual( message_data[ 'student_id' ], self.reviewer.id, "Receiving student id" )
        self.assertEqual( message_data[ 'subject' ], self.activity.email_subject, "Email subject" )
        self.assertTrue( len( message_data[ 'body' ] ) > 0 )

        # todo This relies on another method of the class, would be good to do this independently
        expected_content = self.obj._make_message_content( self.work, None, self.reviewer )
        self.assertEqual( expected_content, message_data[ 'body' ], "Expected message body" )

        # Super important: makes sure going to right person
        # We are sending the content unit out for review, so the
        #  receipient should be the REVIEWER
        self.studentRepo.get_student.assert_called_with(self.reviewer.id )

    @patch('CanvasHacks.Messaging.base.MessageLogger')
    @patch( 'CanvasHacks.Messaging.SendTools.ConversationMessageSender.send' )
    def test_notify( self, sendMock, loggerMock ):
        sendMock.return_value = 'this would be the result of sending'
        self.obj = PeerReviewInvitationMessenger( self.unit, self.studentRepo, self.contentRepo,  self.statusRepo )

        # Call
        result = self.obj.notify( [ self.review_assign ], send=True )

        # Check
        # Intermediate things were called
        self.assertTrue(self.contentRepo.get_formatted_work_by.called, "Content repo method called")
        self.assertTrue(self.studentRepo.get_student.called, "Student repo method called")
        # Returned expected stuff
        self.assertTrue( len( result ) == 1, "Something was returned" )

        # Check that the sender was given the expected content
        sendMock.assert_called()
        kwargs = sendMock.call_args[1 ]
        self.assertEqual( kwargs['student_id'], self.reviewer.id, "Sent to correct student" )
        self.assertEqual(kwargs['subject'], self.activity.email_subject, "Sent with expected subject line")
        d = self.obj._make_template_input( self.work, None, self.reviewer )
        b = REVIEW_NOTICE_TEMPLATE.format(**d)
        self.assertEqual(kwargs['body'], b, "Sent with expected body")

        # Super important: makes sure going to right person
        # We are sending the content unit out for review, so the
        #  receipient should be the REVIEWER
        self.studentRepo.get_student.assert_called_with(self.reviewer.id )

        self.statusRepo.record.assert_called()
        self.statusRepo.record.assert_called_with(self.reviewer.id)

        # self.obj.logger.write.assert_called()



"""
Logging and reporting penalties

Created by adam on 3/18/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass


class PenaltyLogger:

    @staticmethod
    def log(  penalizer ):
        if len( penalizer.penalty_messages ) > 0:
            for msg in penalizer.penalty_messages:
                print(msg)


            # for penalty_dict in penalizer.penalized_records:
            #     due_at = None
            #     try:
            #         due_at = penalizer.due_date
            #     except AttributeError:
            #         # Some penalizers may not have a due date
            #         pass
            #     msg = self._penalty_message( penalty_dict[ 'penalty' ], penalty_dict[ 'record' ], due_at=None )
            #     print(msg)


    # def _penalty_message( self, penalty, row, due_at=None ):
    #     """
    #     Handles printing or logging of penalties applied
    #
    #     # todo enable logging
    #
    #     :param penalty:
    #     :param row:
    #     :return:
    #     """
    #     stem = 'Student #{}: Submitted on {}; was due {}. Penalized {}'
    #     return stem.format( row[ 'student_id' ], row[ 'submitted' ], due_at, penalty )




"""
Objects which define the parameters/ values for the entire unit

Created by adam on 12/24/19
"""
from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Definitions.base import BlockableActivity
from CanvasHacks.Definitions.groups import SkaaReviewGroup, ReviewType
from CanvasHacks.GradingCorrections.penalities import HalfLate, NoLatePenalty
from CanvasHacks.GradingMethods.nonempty import CreditForNonEmptyOLD

__author__ = 'adam'

import re

import pandas as pd
from CanvasHacks.Models.QuizModels import StoredDataFileMixin, QuizDataMixin

if __name__ == '__main__':
    pass


class InitialWork( SkaaReviewGroup, Activity, QuizDataMixin, StoredDataFileMixin ):
    title_base = "Essay"

    # title_base = "Content assignment"
    instructions_filename = 'content-assignment-instructions.txt'
    creation_type = 'assignment'

    # regex = re.compile( r"\bcontent assignment\b" )
    regex = re.compile( r"\bessay\b|\bcontent assignment\b" )

    def __init__( self, **kwargs ):
        self.question_columns = [ ]
        super().__init__( **kwargs )

        # Code for accessing the subsequent unit
        self.access_code_for_next_on = Review
        self.access_code_for_next = None

        # Tools which handle the grade
        self.grace_period = pd.Timedelta( '2 days' )

        self.corrections = [ ]

        # The objects which will be used to penalize late assignments
        self.penalizers = [ HalfLate( self.due_at, self.grace_period ) ]
        # The object which will be used to penalize late assignments

        # The object which will be used to assign the score
        # NB, min_words is 1 for now so as to not create problems with multiple-choice answers
        # This could be fixed in CAN-57
        # The object which will be used to assign the score
        self.grade_methods = [ CreditForNonEmptyOLD( min_words=2, count_stopwords=True ) ]

        try:
            # todo deprecated
            self.penalizer = self.penalizers[ 0 ]
            self.grade_method = self.grade_methods[ 0 ]
        except AttributeError as e:
            print(e)
        #
        # self.composition
        #     (NonEmpty, 0.25),
        #     (WordCount, 0.50),
        #     (ReviewBased, 0.25)
        #
        # ]


class Review( SkaaReviewGroup, Activity, QuizDataMixin, StoredDataFileMixin, ReviewType ):
    """Representation of the peer review component of the
     unit """
    title_base = "Peer review"
    instructions_filename = 'peer-review-instructions.txt'
    creation_type = 'survey'

    regex = re.compile( r"\bpeer review\b" )

    def __init__( self, **kwargs ):
        # Code used to open the review unit
        self.access_code = None

        # todo access_code_for_next_on probably not needed; created without looking at what already have
        self.access_code_for_next_on = MetaReview
        self.access_code_for_next = None

        # Link to the activity_inviting_to_complete on canvas so students can click
        # directly to it
        self.activity_link = None

        self.corrections = [ ]

        # The objects which will be used to penalize late assignments
        self.penalizers = [ NoLatePenalty() ]
        # The object which will be used to penalize late assignments
        # todo deprecated
        # self.penalizer = self.penalizers[ 0 ]

        # The object which will be used to assign the score
        # NB, min_words is 1 for now so as to not create problems with multiple-choice answers
        # This could be fixed in CAN-57
        # The object which will be used to assign the score
        self.grade_methods = [ CreditForNonEmptyOLD( min_words=2, count_stopwords=True ) ]
        # self.grade_method = self.grade_methods[ 0 ]

        try:
            # todo deprecated
            self.penalizer = self.penalizers[ 0 ]
            self.grade_method = self.grade_methods[ 0 ]
        except AttributeError as e:
            print(e)

        super().__init__( **kwargs )
        self.email_intro = "Here is another student's assignment for you to review:"

    @property
    def email_subject( self ):
        """Since unit number will be set after initialization
        we need to do this as a property"""
        return "Unit {} peer-review of content unit".format( self.unit_number )


class MetaReview( SkaaReviewGroup, Activity, BlockableActivity, QuizDataMixin, StoredDataFileMixin, ReviewType ):
    """Representation of the peer review of
    another student's submission"""
    title_base = "Metareview"
    instructions_filename = 'metareview-instructions.txt'
    access_code_for_next_on = None
    creation_type = 'survey'

    regex = re.compile( r"\bmetareview\b" )

    def __init__( self, **kwargs ):
        super().__init__( **kwargs )

        self.email_intro = "Here is the feedback on your assignment:"

        self.corrections = [ ]

        # The objects which will be used to penalize late assignments
        self.penalizers = [ NoLatePenalty() ]
        # The object which will be used to penalize late assignments
        # todo deprecated
        # self.penalizer = self.penalizers[ 0 ]

        # The object which will be used to assign the score
        # NB, min_words is 1 for now so as to not create problems with multiple-choice answers
        # This could be fixed in CAN-57
        # The object which will be used to assign the score
        self.grade_methods = [ CreditForNonEmptyOLD( min_words=2, count_stopwords=True ) ]
        # self.grade_method = self.grade_methods[ 0 ]

        try:
            # todo deprecated
            self.penalizer = self.penalizers[ 0 ]
            self.grade_method = self.grade_methods[ 0 ]
        except AttributeError as e:
            print(e)

    @property
    def email_subject( self ):
        """Since unit number will be set after initialization
        we need to do this as a property"""
        return "Unit {} metareview of peer-review".format( self.unit_number )

# class Assignment( StoreMixin ):
#     """Defines all constant values for the unit"""
#
#     def __init__( self, initial_work: InitialWork, review: Review, meta_review: MetaReview, **kwargs ):
#         """
#         :param initial_work:
#         :param review:
#         :param meta_review:
#         """
#         self.initial_work = initial_work
#         self.meta_review = meta_review
#         self.review = review
#
#         self.handle_kwargs( kwargs )



"""
This contains the tools for finding reports in the downloads folder
and moving them to the appropriate locations

Created by adam on 3/24/20
"""
__author__ = 'adam'

import os
import re
import shutil
from functools import wraps
from CanvasHacks import environment as env
from CanvasHacks.PeerReviewed.Definitions import Unit
from CanvasHacks.TimeTools import current_utc_timestamp, timestamp_for_unique_filenames


UNIT_RX = re.compile( r"(\bunit\b) (\d+)" )

DOWNLOAD_FOLDER = "{}/Downloads".format(env.ROOT)


def normalize_filename( f ):
    @wraps( f )
    def d( *args, **kwargs ):
        fname = args[ 0 ].strip().lower()
        return f( fname, **kwargs )

    return d


@normalize_filename
def is_csv( filename ):
    """
    Returns true if filename ends in csv
    :param filename:
    :return:
    """
    # f = filename.strip().lower()
    return filename[ -4: ] == '.csv'


@normalize_filename
def is_report( filename ):
    """
    Returns true if seems to be a report, false otherwise
    :param filename:
    :return:
    """
    # f = filename.strip().lower()

    return UNIT_RX.search( filename ) is not None


@normalize_filename
def get_unit_number( filename ):
    # f = filename.strip().lower()
    match = UNIT_RX.search( filename )
    # start = match.span[0]
    # stop = match.span[1]
    return int( match.group( 2 ) )


@normalize_filename
def get_activity_type( filename ):
    for c in Unit.component_types:
        if c.is_activity_type( filename ):
            return c


def report_file_iterator( folderPath, exclude=[ ] ):
    exclude = exclude if any( exclude ) else [ '.DS_Store' ]
    for root, dirs, files in os.walk( folderPath ):
        for name in files:
            if name not in exclude:
                if is_csv( name ) and is_report( name ):
                    path = os.path.join( root, name )
                    d = {
                        'unit_number': get_unit_number( name ),
                        'file_name': name,
                        'path': path,
                        'activity': get_activity_type( name )
                    }
                    yield d


def file_reports(download_folder_path, unit_start=1, unit_stop=6):
    """
    Moves reports found in the download folder to the
    correct folder
    :param download_folder_path:
    :return:
    """

    units = { u:  Unit( env.CONFIG.course, u ) for u in range(unit_start, unit_stop + 1) }
    fiter = report_file_iterator( download_folder_path )
    moved_files = []
    try:
        while True:
            f = next(fiter)
            if f['activity'] is not None:
                unit = units.get(f['unit_number'])
                activity = unit.get_by_class(f['activity'])
                src = f['path']

                # Rename with timestamp so no collisions
                # THIS DOES NOT GUARANTEE THAT THE NEWEST FILE CAN BE DETERMINED BY
                # TIMESTAMP!
                to = "{}/{} {}".format(activity.folder_path, timestamp_for_unique_filenames(), f['file_name'])
                shutil.move(src, to)
                moved_files.append({'from': src, 'to': to})

    except StopIteration:
        for f in moved_files:
            print("\n----\n FROM: {from} \n TO: {to}".format(**f))
        return moved_files


if __name__ == '__main__':
    DOWNLOAD_FOLDER = "{}/Downloads".format( env.ROOT )

    file_reports(DOWNLOAD_FOLDER)



"""
Created by adam on 3/16/20
"""
__author__ = 'adam'

from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.GradingAnalyzers.nonempty import NonEmpty
from CanvasHacks.GradingMethods.review import ReviewBased
from CanvasHacks.Models.model import Model
from CanvasHacks.Text.stats import WordCount

if __name__ == '__main__':
    pass

class ActivityPoints(Model):

    def __init__(self, activity: Activity, total_points):
        self.activity = activity
        self.total_points = total_points

        self.composition = [
            (NonEmpty, 0.25),
            (WordCount, 0.50),
            (ReviewBased, 0.25)

        ]


"""
Created by adam on 3/9/20
"""
__author__ = 'adam'
# first so will set globals
from CanvasHacks.Definitions.discussion import DiscussionForum
from TestingBase import TestingBase

from unittest.mock import MagicMock, create_autospec

from CanvasHacks.GradingHandlers.discussion import DiscussionForumGrader
from CanvasHacks.Repositories.discussions import DiscussionRepository

# from CanvasHacks.Repositories.discussions import DiscussionRepository
from factories.ModelFactories import student_factory

if __name__ == '__main__':
    pass


class TestDiscussionForumGrader( TestingBase ):

    def make_student_post( self, student, is_blank=False ):

        return {
            'student_id': student.student_id,
            'student_name': student.name,
            'text': '' if is_blank else self.fake.paragraph()
        }

    def make_post_data( self ):
        if len( self.students ) == 0:
            self.create_new_and_preexisting_students()

        self.data = [ ]
        self.number_w_required_or_more = 0
        for s in self.students:
            posts = self.fake.random.randint( 0, 50 )
            for _ in range( 0, posts ):
                self.data.append( self.make_student_post( s ) )

            if posts >= self.number_required:
                self.number_w_required_or_more += 1

    def setUp( self ):
        self.config_for_test()
        self.create_new_and_preexisting_students()

        self.number_required = self.fake.random.randint( 2, 30 )
        self.expected_per_post = 100 / self.number_required
        self.make_post_data()
        # use real graders
        self.activity = DiscussionForum()
        # self.work_repo = MagicMock( data=self.data,  )
        # creates problems when runnning offline
        self.work_repo = create_autospec( DiscussionRepository, data=self.data, activity=self.activity )

        self.grade_func = MagicMock()

    def test_initialize( self ):

        obj = DiscussionForumGrader( self.work_repo, num_posts_required=self.number_required )

        # check
        self.assertEqual( obj.credit_per_post, self.expected_per_post, "Correct credit per post set" )

    def test__prepare_results( self ):
        self.skipTest('fuck you testing lib')
        num_required = 2
        obj = DiscussionForumGrader( self.work_repo, num_posts_required=num_required )

        # below required
        expected = [ (1, 50), (2, 0) ]
        obj._raw = [ (1, 50), (2, 0) ]
        obj._prepare_results()

        self.assertEqual( obj.graded, expected, "Posts under required are untouched" )

        # at required
        expected = [ (1, 100), (2, 100) ]
        obj._raw = [ (1, 50), (1, 50), (2, 50), (2, 50) ]
        obj._prepare_results()

        self.assertEqual( obj.graded, expected, "Posts equal to required are summed" )

        # over required
        expected = [ (1, 100), (2, 100) ]
        obj._raw = [ (1, 50), (1, 50), (1, 50), (2, 50), (2, 50), (2, 50) ]
        obj._prepare_results()

        self.assertEqual( obj.graded, expected, "Posts over required are reduced to 100" )

    def test__calc_scores_blank( self ):
        student = self.students[ 0 ]
        data = self.make_student_post( student, is_blank=True )

        work_repo = create_autospec( DiscussionRepository, data=[ data ], activity=self.activity  )
        obj = DiscussionForumGrader( work_repo, num_posts_required=self.number_required )

        # call
        obj._calc_scores()

        # check
        self.assertEqual( obj._raw[ 0 ], (student.student_id, 0), "Receives 0 when blank" )

    def test__calc_scores( self ):
        student = self.students[ 0 ]
        data = self.make_student_post( student )

        work_repo = create_autospec( DiscussionRepository, data=[ data ], activity=self.activity  )
        obj = DiscussionForumGrader( self.work_repo )

        # call
        obj._calc_scores()

        # check
        self.assertEqual( obj._raw[ 0 ], (student.student_id, obj.credit_per_post),
                          "Receives credit_per_post when not blank" )

    def test_grade_eq_to_expected( self ):
        # student who does exactly what's required
        student = student_factory()
        student_posts = [ self.make_student_post(student) for _ in range(0, self.number_required)]

        work_repo = create_autospec( DiscussionRepository, data=student_posts, activity=self.activity   )
        obj = DiscussionForumGrader( work_repo, num_posts_required=self.number_required )

        # call
        obj.grade()

        # check
        self.assertEqual(len(obj.graded), 1, "correct number of students in output")
        self.assertEqual((student.student_id, 100), obj.graded[0], "Correct result stored in obj.graded")

    def test_grade_greater_than_expected( self ):
        # student who does more than what's required
        student = student_factory()
        student_posts = [ self.make_student_post(student) for _ in range(0, self.number_required + 10)]

        work_repo = create_autospec( DiscussionRepository, data=student_posts, activity=self.activity  )
        obj = DiscussionForumGrader( work_repo, num_posts_required=self.number_required )

        # call
        obj.grade()

        # check
        self.assertEqual( len( obj.graded), 1, "correct number of students in output" )
        self.assertEqual((student.student_id, 100), obj.graded[0], "Correct result stored in obj.graded")

    def test_grade_less_than_expected( self ):
        # student who does less than required
        num_posts = self.number_required - 1

        student = student_factory()
        student_posts = [ self.make_student_post( student ) for _ in range( 0,  num_posts) ]

        work_repo = create_autospec( DiscussionRepository, data=student_posts, activity=self.activity  )
        obj = DiscussionForumGrader( work_repo, num_posts_required=self.number_required )

        # call
        obj.grade()

        # check
        self.assertEqual( len( obj.graded), 1, "correct number of students in output" )
        expected_score =round( num_posts * self.expected_per_post)
        self.assertAlmostEqual(expected_score, obj.graded[0][1], places=3, msg="Correct result stored in obj.graded")






"""
Created by adam on 2/18/20
"""

__author__ = 'adam'

from CanvasHacks.GradingHandlers.base import IGrader

__author__ = 'adam'

from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.GradingCorrections.penalities import IPenalizer


class JournalGrader( IGrader ):
    """Handles grading weekly journals"""
    grade_method: IGradingMethod
    penalizer: IPenalizer

    def __init__( self, work_repo, **kwargs ):
        """
        :param work_repo: Content repository with journal data
        :param kwargs:
        """
        self.work_repo = work_repo
        super().__init__( **kwargs )
        self.penalizer = self.activity.penalizer
        self.grade_method = self.activity.grade_method
        self.graded = [ ]

    def grade( self ):
        """Assigns a provisional grade for the journal
        Will return as a list of tuples
        todo: Add logging of details of how grade assigned

       Determines how much credit potentially late credit/no credit
        assignments should receive.
        Created in CAN-24
        """
        for submission in self.work_repo.data:
            if submission.body is not None:
                score = self.grade_method.grade( submission.body )
                if score:
                    score = self.penalizer.get_penalized_score( submission.submitted_at, score, record=submission )
                    self.graded.append( (submission, int( score )) )

        self.report_late_penalties()

        return self.graded

    def report_late_penalties( self ):
        # Report late penalties
        if len( self.penalizer.penalized_records ) > 0:
            for penalty_dict in self.penalizer.penalized_records:
                self._penalty_message(penalty_dict['penalty'], penalty_dict['record'])


    def _penalty_message( self, penalty, record ):
        """
        Handles printing or logging of penalties applied

        # todo enable logging

        :param penalty:
        :param row: A Submission object
        :return:
        """
        stem = 'Student #{}: Submitted on {}; was due {}. Penalized {}'
        try:
            return stem.format( record.student_id, record.submitted, self.activity.due_at, penalty )
        except (TypeError, AttributeError):
            # Will hit this if a Submission object
            return stem.format( record.user_id, record.submitted_at, self.activity.due_at, penalty )


if __name__ == '__main__':
    pass



"""
Created by adam on 2/5/19
"""
__author__ = 'adam'

from IPython.display import display
from ipywidgets import widgets

from CanvasHacks.Widgets.InputFields import make_course_ids_input, make_canvas_token_input

TOKEN_INSTRUCTIONS = """
    <div>
        <div>
            <p>You'll need to generate a token for the Canvas API. The token replaces your username/password when making requests to Canvas.</p> 
            <p>Please record the token in a secure place. This program will not save it; you will have to re-enter it every time you use this tool.</p>

            <p>First log into Canvas. Then: </p>
            <blockquote>
                <p>Click the "profile" link in the top right menu bar, or navigate to /profile</p>

                <p>Under the "Approved Integrations" section, click the button to generate a new access token.</p>

            <p>Once the token is generated, you cannot view it again, and you'll have to generate a new token if you forget
                it.
                Remember that access tokens are password equivalent, so keep it secret (ed: keep it safe!).</p>

                <p><a href="https://canvas.instructure.com/doc/api/file.oauth.html#manual-token-generation">https://canvas.instructure.com/doc/api/file.oauth.html#manual-token-generation</a>
                </p>
            </blockquote>
        </div>
        <div>
            <h5> Paste your token into the box below. You don't have to click anything, its value will be read when
                needed</h5>
        </div>
    </div>
    """

SECTION_INSTRUCTIONS = """
    <div>
        <p>You will also need the id number(s) of the courses you wish to grade. You can find them by logging into canvas
            and going to the course page. The number you are looking for will be in the url, immediately after
            '/courses/'. For example:</p>
        <blockquote>https://canvas.csun.edu/courses/<b>12345</b></blockquote>
        <p>Write this number down somewhere secure so that you can re-enter it next time you use the
            program.</p>

        <h5>Add the course ids one-at-a-time using the box and buttons below</h5>
    </div>
    """


def token_instructions_and_input():
    instruct = widgets.HTML( value=TOKEN_INSTRUCTIONS )
    display( instruct )
    make_canvas_token_input()


def section_instructions_and_input():
    display( widgets.HTML( SECTION_INSTRUCTIONS ) )
    make_course_ids_input()


def make_credentials_input():
    token_instructions_and_input()
    section_instructions_and_input()


if __name__ == '__main__':
    pass



"""
Created by adam on 3/16/20
"""
__author__ = 'adam'

import re

import pandas as pd

from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.GradingCorrections.penalities import HalfLate
from CanvasHacks.GradingMethods.nonempty import CreditForNonEmptyOLD

if __name__ == '__main__':
    pass


class Journal( Activity ):
    """Representation of a journal unit.
    Not related to assignments within a Unit
    """
    title_base = "Journal"

    regex = re.compile( r"\bjournal\b" )
    creation_type = 'assignment'

    def __init__( self, **kwargs ):
        self.grace_period = pd.Timedelta( '2 days' )
        super().__init__( **kwargs )

        self.corrections = [ ]

        # The objects which will be used to penalize late assignments
        self.penalizers = [HalfLate( self.due_at, self.grace_period )]
        # The object which will be used to penalize late assignments
        # todo deprecated
        self.penalizer = self.penalizers[ 0 ]

        # The object which will be used to assign the score
        # NB, min_words is 1 for now so as to not create problems with multiple-choice answers
        # This could be fixed in CAN-57
        # The object which will be used to assign the score
        self.grade_methods = [ CreditForNonEmptyOLD( min_words=2, count_stopwords=True ) ]
        self.grade_method = self.grade_methods[ 0 ]




"""
Created by adam on 2/24/20
"""
import pandas as pd

__author__ = 'adam'

if __name__ == '__main__':
    pass


def process_work( work_frame, submissions_frame ):
    try:
        v = work_frame[ 'student_id' ]
    except KeyError:
        work_frame.rename( { 'id': 'student_id' }, axis=1, inplace=True )
    # merge it with matching rows from the submissions frame
    frame = pd.merge( work_frame, submissions_frame, how='left', on=[ 'student_id', 'attempt' ] )
    try:
        # Try to sort it on student names if possible
        frame.set_index( 'name', inplace=True )
        frame.sort_index( inplace=True )
    except KeyError:
        pass
    return frame


def remove_non_final_attempts( frame ):
    """Can only be ran after submission has been added?"""
    frame.dropna( subset=[ 'submission_id' ], inplace=True )


def make_drop_list( columns ):
    """The canvas exports will have some annoying fields
        These should be added to the droppable list
        If there are columns with a common initial string (e.g., 1.0, 1.0.1, ...) just
        add the common part
    """

    droppable = [ '1.', 'Unnamed' ]
    to_drop = ['n correct', 'n incorrect' ]
    for c in columns:
        try:
            if float( c ):
                to_drop.append( c )
        except ValueError:
            for d in droppable:
                if c[ :len( d ) ] == d:
                    to_drop.append( c )
    return to_drop


def check_responses( row, question_columns ):
    score = 0
    for c in question_columns:
        try:
            if pd.isnull( row[ c ] ):
                raise Exception

            # No credit for 1 word answers
            if len( row[ c ] ) < 2:
                raise Exception

            # If we made it past the tests, increment the score
            score += 1
        except Exception:
            pass

    return score


def add_graded_total_field( frame, question_columns ):
    frame[ 'graded_total' ] = frame.apply( lambda r: check_responses( r, question_columns ), axis=1 )


def drop_columns_from_frame( frame ):
    to_drop = make_drop_list( frame.columns )
    frame.drop( to_drop, axis=1, inplace=True )
    print( "Removed: ", to_drop )


"""
Errors and special cases that arise when trying to pair up
 students for peer review.
Created by adam on 1/23/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass

class AllAssigned(Exception):
    """Raised when every student who has submitted new has
    previously been assigned a partner. This could arise
    if they resubmitted the unit after the deadline and
    were thus treated as new
    """
    def __init__(self, submitters):
        self.submitters = submitters

class AlreadyAssigned( Exception ):
    """Raised when the student whose work is being considered for unit
    has already had someone else assigned to grade it
    """
    pass


class SubmissionIncomplete( Exception ):
    """Raised when the submission object being considered for
    review assignent has not yet been completed"""
    pass


class NoAvailablePartner( Exception ):
    """Raised when only 1 new person has submitted and thus
    there is no one to review them"""
    def __init__(self, submitters):
        self.submitters = submitters

class NoReviewPairingsLoaded(Exception):
    pass

class NoReviewPairingFound(Exception):
    """Raised when no review assignment could be found
    for the specified student.

    This can happen when someone turns in the peer review without
    turning in the essay
    """
    def __init__(self, student_id):
        self.student_id = student_id
        print("No review pairing found for student {}".format(self.student_id))



"""
Created by adam on 3/10/20
"""
__author__ = 'adam'

from CanvasHacks.GradingHandlers.base import IGrader
from CanvasHacks.GradingHandlers.quiz import QuizGrader
from CanvasHacks.Repositories.interfaces import ISubmissionRepo
from CanvasHacks.Repositories.quizzes import QuizRepository

if __name__ == '__main__':
    pass


class ReviewGrader(QuizGrader):
    """Handles grading review assignments
    These will be essentially quiz assignments but we may
    want to handle the multiple choice differently
    todo Currently this clones QuizGrader consider updating all of this in CAN-57
    """

    def __init__( self, work_repo: QuizRepository, submission_repo: ISubmissionRepo, **kwargs ):
        """
        :param work_repo: Content repository with review data
        :param kwargs:
        """
        self.work_repo = work_repo
        self.submission_repo = submission_repo

        # super().__init__( **kwargs )

        self.penalizer = self.activity.penalizer
        self.grade_method = self.activity.grade_method

        self.graded = [ ]

    def _grade_row( self, row , **kwargs):
        """Grades a row

        NB, requires that the non-graded attempts be
        filtered out before passing in

        todo test whether the problem with non-graded people getting zeros is caused here

        :param row: pd.DataFrame row
        :param kwargs:
        :return: Dictionary formatted for uploading
        """
        # used for computing penalty
        total_score = 0
        questions = { }

        # Grade on emptiness
        # todo CAN-57 Consider implementing separate graders for text and multiple choice columns. That will allow to require more words in the text columns without hosing the 1-2 word multiple choice answers
        for qid, column_name in self.work_repo.question_columns:
            content = row[ column_name ]
            pts = self._get_score( content, **kwargs )
            # Quiz type things require the scores to be uploaded
            # for each question.
            # Thus we grade each question and then penalize the
            # overall score with fudge points if necessary
            questions[ qid ] = { 'score': pts }
            total_score += pts

        # Compute penalty if needed
        # Will be 0 if not docking for lateness
        # Records of penalty will be stored on self.penalizer.penalized_records
        fudge_points = self.penalizer.get_fudge_points(row['submitted'], total_score, row)

        out = self._make_graded_row_output(row, questions, fudge_points)

        return out